import matplotlib
import matplotlib.pyplot as plt
import matplotlib.path as mpath
import matplotlib.patches as mpatches
import numpy as np
import math
from collections import OrderedDict
from lib.common.genome.genome_feature2 import read_gff_file
from interlap import InterLap
from lib.common.math.bin_and_window import split_sequence_to_bins
from lib.common.genome.seq_base import read_fasta_by_faidx
from lib.common.math.interval import merge_intervals, sum_interval_length


class CirclePlot(object):
    def __init__(self, **kwargs):
        # set default parameter
        self.chr_dict = OrderedDict()

        # figure parameter
        self.init_figure_parameter = {
            "figure_size": (20, 20),
            "zero_location": "E",
            "ylim": (0, 5000),
            "direction": -1,
            "zero_location_gap": (2*math.pi/365)*15,
            "interspace_ratio": 0.1,
            "show_polar_axis": False,
        }

        for k in kwargs:
            if k in self.init_figure_parameter:
                self.init_figure_parameter[k] = kwargs[k]

        # build figure and axes
        self.figure = plt.figure(
            figsize=self.init_figure_parameter["figure_size"])
        self.ax = plt.subplot(111, polar=True)
        self.ax.set_theta_zero_location(
            self.init_figure_parameter["zero_location"])
        self.ax.set_theta_direction(self.init_figure_parameter["direction"])
        self.ax.set_ylim(*self.init_figure_parameter["ylim"])

        if not self.init_figure_parameter["show_polar_axis"]:
            # remove polar axis
            self.ax.spines['polar'].set_visible(False)
            self.ax.xaxis.set_ticks([])
            self.ax.xaxis.set_ticklabels([])
            self.ax.yaxis.set_ticks([])
            self.ax.yaxis.set_ticklabels([])

    def load_chr_info(self, chr_id, length, **kwargs):
        self.chr_dict[chr_id] = {}
        self.chr_dict[chr_id]["length"] = length
        self.chr_dict[chr_id]["features"] = []

        for k in ["bottom", "height", "facecolor", "edgecolor", "linewidth"]:
            if k in kwargs:
                self.chr_dict[chr_id][k] = kwargs[k]
            else:
                self.chr_dict[chr_id][k] = None

        self.sum_chr_length = sum([self.chr_dict[i]["length"]
                                   for i in self.chr_dict])
        sum_interspace = self.sum_chr_length * \
            self.init_figure_parameter["interspace_ratio"]
        self.interspace_num = len(self.chr_dict)
        self.interspace_length = sum_interspace/self.interspace_num
        sum_length = self.sum_chr_length + sum_interspace

        sum_theta = 2*math.pi - self.init_figure_parameter["zero_location_gap"]
        self.the_vs_len = sum_theta/sum_length
        self.interspace_theta = self.interspace_length * self.the_vs_len
        self.theta_start = (
            self.init_figure_parameter["zero_location_gap"] + self.interspace_theta) / 2

        tmp_start = self.theta_start
        for chr_id in self.chr_dict:
            theta_s = tmp_start
            theta_e = theta_s + \
                self.chr_dict[chr_id]['length'] * self.the_vs_len
            self.chr_dict[chr_id]['theta'] = (theta_s, theta_e)
            self.chr_dict[chr_id]['theta_length'] = theta_e - theta_s
            tmp_start = theta_e + self.interspace_theta

    def chr_site_to_theta(self, chr_id, site):
        chr_theta_start = self.chr_dict[chr_id]['theta'][0]
        return chr_theta_start + site * self.the_vs_len

    def chrs_plot(self, **kwargs):
        # parameter
        # default parameter

        chr_box_parameter = {
            "bottom": 3500,
            "height": 250,
            "facecolor": "none",
            "edgecolor": "#000000",
            "linewidth": 1.0,
            "ticker": True,
            "ticker_base": 10000000,
            "ticker_length": 10,
            "ticker_label": True,
            "ticker_label_length": 100,
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in chr_box_parameter:
                chr_box_parameter[k] = kwargs[k]

        # add chr box
        for chr_id in self.chr_dict:
            pos = self.chr_dict[chr_id]["theta"][0]
            width = self.chr_dict[chr_id]["theta_length"]

            if self.chr_dict[chr_id]["height"] is None:
                height = chr_box_parameter["height"]
            else:
                height = self.chr_dict[chr_id]["height"]

            if self.chr_dict[chr_id]["bottom"] is None:
                bottom = chr_box_parameter["bottom"]
            else:
                bottom = self.chr_dict[chr_id]["bottom"]

            if self.chr_dict[chr_id]["facecolor"] is None:
                facecolor = chr_box_parameter["facecolor"]
            else:
                facecolor = self.chr_dict[chr_id]["facecolor"]

            if self.chr_dict[chr_id]["edgecolor"] is None:
                edgecolor = chr_box_parameter["edgecolor"]
            else:
                edgecolor = self.chr_dict[chr_id]["edgecolor"]

            if self.chr_dict[chr_id]["linewidth"] is None:
                linewidth = chr_box_parameter["linewidth"]
            else:
                linewidth = self.chr_dict[chr_id]["linewidth"]

            self.chr_dict[chr_id]["bar"] = self.ax.bar(
                [pos], [height], bottom=bottom, width=width, facecolor=facecolor, linewidth=linewidth, edgecolor=edgecolor, align="edge")

        # add chr tick and tick label
        for chr_id in self.chr_dict:
            chr_theta_start = self.chr_dict[chr_id]["theta"][0]
            length = self.chr_dict[chr_id]["length"]

            if self.chr_dict[chr_id]["height"] is None:
                height = chr_box_parameter["height"]
            else:
                height = self.chr_dict[chr_id]["height"]

            if self.chr_dict[chr_id]["bottom"] is None:
                bottom = chr_box_parameter["bottom"]
            else:
                bottom = self.chr_dict[chr_id]["bottom"]

            if self.chr_dict[chr_id]["linewidth"] is None:
                linewidth = chr_box_parameter["linewidth"]
            else:
                linewidth = self.chr_dict[chr_id]["linewidth"]

            tick_theta_positions_list = []
            for i in range(int(length/chr_box_parameter["ticker_base"])+1):
                tick_theta_positions_list.append(
                    chr_theta_start + i * chr_box_parameter["ticker_base"] * self.the_vs_len)

            # add tick
            tick_bottom = bottom + height
            if chr_box_parameter["ticker"]:
                for tick_theta in tick_theta_positions_list:
                    self.ax.plot([tick_theta, tick_theta], [
                        tick_bottom, tick_bottom+chr_box_parameter["ticker_length"]], color="k", linewidth=linewidth)

            # add tick label
            if chr_box_parameter["ticker_label"]:
                tick_label_bottom = tick_bottom + \
                    chr_box_parameter["ticker_length"] + \
                    chr_box_parameter["ticker_label_length"]
                num = 0
                for tick_theta in tick_theta_positions_list:
                    label_text = str(int(num / 1000000))

                    if tick_theta <= math.pi:
                        lab = self.ax.text(tick_theta, tick_label_bottom, label_text, ha='center',
                                        va='center', rotation=np.rad2deg(2*math.pi - (tick_theta - 0.5*math.pi)))
                    else:
                        lab = self.ax.text(tick_theta, tick_label_bottom, label_text, ha='center',
                                        va='center', rotation=np.rad2deg(math.pi - (tick_theta - 0.5*math.pi)))

                    num += chr_box_parameter["ticker_base"]

    def add_chrs_name(self, bottom, **kwargs):
        chrs_name_parameter = {
            "fontsize": "xx-large",
            "color": "#000000",
            "family": 'Arial Unicode MS'
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in chrs_name_parameter:
                chrs_name_parameter[k] = kwargs[k]

        for chr_id in self.chr_dict:
            chrs_name_theta = (self.chr_dict[chr_id]["theta"][1] - self.chr_dict[chr_id]
                               ["theta"][0]) / 2 + self.chr_dict[chr_id]["theta"][0]

            if self.init_figure_parameter["zero_location"] == 'E':
                rotation_now=np.rad2deg(math.pi - (chrs_name_theta - 0.5*math.pi))
            if self.init_figure_parameter["zero_location"] == 'N':
                rotation_now=np.rad2deg(math.pi - (chrs_name_theta - math.pi))
            if self.init_figure_parameter["zero_location"] == 'S':
                rotation_now=np.rad2deg(math.pi - (chrs_name_theta - 1.5*math.pi))
            if self.init_figure_parameter["zero_location"] == 'w':
                rotation_now=np.rad2deg(math.pi - (chrs_name_theta - 0*math.pi))

            lab = self.ax.text(chrs_name_theta, bottom, chr_id, ha='center',
                            va='center', rotation=rotation_now, fontsize=chrs_name_parameter['fontsize'], color=chrs_name_parameter['color'], family=chrs_name_parameter['family'])

    def heatmap_plot(self, data, bin_length, bottom, height, min_plot=None, max_plot=None, cmap=None):
        """
        data = {
            'Chr1' : {0:100,1:20}
            }
        }
        0,1 is index for bins, and 100,20 is value to plot

        should load_chr_info firstly
        """
        bar_width = bin_length * self.the_vs_len

        num_list = []
        for i in data:
            for j in data[i]:
                num_list.append(data[i][j])

        if max_plot:
            max_value = max_plot
        else:
            max_value = max(num_list)

        if min_plot:
            min_value = min_plot
        else:
            min_value = min(num_list)

        for chr_id in self.chr_dict:
            theta = self.chr_dict[chr_id]['theta']
            chr_length = self.chr_dict[chr_id]['length']
            bar_start = []
            theta_start = theta[0]
            for i, s, e in split_sequence_to_bins(chr_length, bin_length, 1):
                bar_start.append(theta_start)
                theta_start += bar_width
            bars = self.ax.bar(bar_start, height=[
                               height] * len(bar_start), bottom=bottom, align="edge", width=bar_width)

            for i, bar in enumerate(bars):
                if i in data[chr_id]:
                    value = data[chr_id][i]
                else:
                    value = min_value

                if value >= max_value:
                    value = max_value
                elif value <= min_value:
                    value = min_value

                bar.set_facecolor(cmap(value/(max_value - min_value)))
                bar.set_linewidth(0.0)

    def line_plot(self, data, bin_length, bottom, height, **kwargs):
        """
        y_axis is a function
        such as:
        y_axis = lambda x:"%.0f%%" % (x * 100)
        """

        line_plot_para = {
            "min_plot": None,
            "max_plot": None,
            "linewidth": 0.5,
            "edgecolor": "k",
            "facecolor": "#5b7281",
            "centripetal": False,
            "fill": True,
            "y_axis": None,
            "axis_linewidth": 1,
            "axis_color": "k",
            "axis_ticker_num": 5,
            "ticker_length": (2*math.pi/365)*0.5,
            "ticker_label_adjust": 5
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in line_plot_para:
                line_plot_para[k] = kwargs[k]

        bar_width = bin_length * self.the_vs_len

        num_list = []
        for i in data:
            for j in data[i]:
                num_list.append(data[i][j])

        if line_plot_para["max_plot"]:
            max_value = line_plot_para["max_plot"]
        else:
            max_value = max(num_list)

        if line_plot_para["min_plot"]:
            min_value = line_plot_para["min_plot"]
        else:
            min_value = min(num_list)

        he_vs_val = height / abs(max_value - min_value)

        for chr_id in self.chr_dict:
            theta = self.chr_dict[chr_id]['theta']
            chr_length = self.chr_dict[chr_id]['length']

            position_list = []
            value_list = []

            theta_start = theta[0]
            for i, s, e in split_sequence_to_bins(chr_length, bin_length, 1):
                position_list.append(theta_start)

                if i in data[chr_id]:
                    value = data[chr_id][i]
                else:
                    value = min_value

                if value >= max_value:
                    value = max_value

                elif value <= min_value:
                    value = min_value

                if line_plot_para["centripetal"]:
                    plot_value = bottom - (value - min_value) * he_vs_val
                else:
                    plot_value = bottom + (value - min_value) * he_vs_val

                value_list.append(plot_value)

                theta_start += bar_width

            if line_plot_para["fill"] == False:
                self.ax.plot(position_list, value_list,
                             color=line_plot_para["edgecolor"], linewidth=line_plot_para["linewidth"])

            elif line_plot_para["fill"] == True:
                if bottom >= 0:
                    self.ax.fill_between(position_list, value_list, bottom,
                                         facecolor=line_plot_para["facecolor"], linewidth=line_plot_para["linewidth"], edgecolor=line_plot_para["edgecolor"])

        # add y-axis
        if line_plot_para["y_axis"]:
            y_axis_theta = self.chr_dict[list(
                self.chr_dict.keys())[0]]['theta'][0] - (2*math.pi/365)*0.5

            # add spine
            spine_start = bottom
            if line_plot_para["centripetal"]:
                spine_end = bottom - height
            else:
                spine_end = bottom + height

            self.ax.plot([y_axis_theta, y_axis_theta], [
                         spine_start, spine_end], color="k", linewidth=line_plot_para["axis_linewidth"])

            # add tick
            tick_height_list = [spine_start, spine_end]
            for i in range(line_plot_para["axis_ticker_num"]):
                tick_height_list.append(
                    spine_start + i * (spine_end - spine_start) / line_plot_para["axis_ticker_num"])

            for tick_theta in tick_height_list:
                self.ax.plot([y_axis_theta, y_axis_theta - line_plot_para["ticker_length"]], [
                    tick_theta, tick_theta], color="k", linewidth=line_plot_para["axis_linewidth"])

            # add tick label
            tick_label_theta = y_axis_theta - \
                line_plot_para["ticker_label_adjust"] * \
                line_plot_para["ticker_length"]
            for tick_label_height, tick_label_value in [(spine_start, min_value), (spine_end, max_value)]:

                lab = self.ax.text(tick_label_theta, tick_label_height, line_plot_para["y_axis"](tick_label_value), ha='center',
                                   va='center', rotation=np.rad2deg(2*math.pi - (tick_label_theta - 0.5*math.pi)))

    def bar_plot(self, data, bin_length, bottom, height, **kwargs):
        """
        y_axis is a function
        such as:
        y_axis = lambda x:"%.0f%%" % (x * 100)
        """

        bar_plot_para = {
            "min_plot": None,
            "max_plot": None,
            "linewidth": 0.0,
            "edgecolor": "k",
            "facecolor": "#5b7281",
            "centripetal": False,
            "y_axis": None,
            "axis_linewidth": 1,
            "axis_color": "k",
            "axis_ticker_num": 5,
            "ticker_length": (2*math.pi/365)*0.5,
            "ticker_label_adjust": 5
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in bar_plot_para:
                bar_plot_para[k] = kwargs[k]

        bar_width = bin_length * self.the_vs_len

        num_list = []
        for i in data:
            for j in data[i]:
                num_list.append(data[i][j])

        if bar_plot_para["max_plot"]:
            max_value = bar_plot_para["max_plot"]
        else:
            max_value = max(num_list)

        if bar_plot_para["min_plot"]:
            min_value = bar_plot_para["min_plot"]
        else:
            min_value = min(num_list)

        he_vs_val = height / abs(max_value - min_value)

        for chr_id in self.chr_dict:
            theta = self.chr_dict[chr_id]['theta']
            chr_length = self.chr_dict[chr_id]['length']

            position_list = []
            value_list = []

            theta_start = theta[0]
            for i, s, e in split_sequence_to_bins(chr_length, bin_length, 1):
                position_list.append(theta_start)

                if i in data[chr_id]:
                    value = data[chr_id][i]
                else:
                    value = min_value

                if value >= max_value:
                    value = max_value

                elif value <= min_value:
                    value = min_value

                if bar_plot_para["centripetal"]:
                    plot_value = - (value - min_value) * he_vs_val
                else:
                    plot_value = + (value - min_value) * he_vs_val

                value_list.append(plot_value)

                theta_start += bar_width

            self.ax.bar(position_list, value_list, bottom=bottom, width=[bar_width] * len(position_list),
                        facecolor=bar_plot_para["facecolor"], linewidth=bar_plot_para["linewidth"], edgecolor=bar_plot_para["edgecolor"])

        # add y-axis
        if bar_plot_para["y_axis"]:
            y_axis_theta = self.chr_dict[list(
                self.chr_dict.keys())[0]]['theta'][0] - (2*math.pi/365)*0.5

            # add spine
            spine_start = bottom
            if bar_plot_para["centripetal"]:
                spine_end = bottom - height
            else:
                spine_end = bottom + height

            self.ax.plot([y_axis_theta, y_axis_theta], [
                         spine_start, spine_end], color="k", linewidth=bar_plot_para["axis_linewidth"])

            # add tick
            tick_height_list = [spine_start, spine_end]
            for i in range(bar_plot_para["axis_ticker_num"]):
                tick_height_list.append(
                    spine_start + i * (spine_end - spine_start) / bar_plot_para["axis_ticker_num"])

            for tick_theta in tick_height_list:
                self.ax.plot([y_axis_theta, y_axis_theta - bar_plot_para["ticker_length"]], [
                    tick_theta, tick_theta], color="k", linewidth=bar_plot_para["axis_linewidth"])

            # add tick label
            tick_label_theta = y_axis_theta - \
                bar_plot_para["ticker_label_adjust"] * \
                bar_plot_para["ticker_length"]
            for tick_label_height, tick_label_value in [(spine_start, min_value), (spine_end, max_value)]:

                lab = self.ax.text(tick_label_theta, tick_label_height, bar_plot_para["y_axis"](tick_label_value), ha='center',
                                   va='center', rotation=np.rad2deg(2*math.pi - (tick_label_theta - 0.5*math.pi)))

    def link_plot(self, start, end, bottom, **kwargs):
        """
        start = ('Chr1', 1000000, 2000000)
        end = ('Chr2', 5000000, 7000000)
        bottom = 1000
        """

        link_plot_para = {
            "bottom": bottom,
            "facecolor": "#5b7281",
            "linewidth": 0.0,
            "edgecolor": "k",
            "alpha": 0.5,
            "center": 0,
            "end_bottom": None,
        }

        for k in kwargs:
            if k in link_plot_para:
                link_plot_para[k] = kwargs[k]

        ss = self.chr_site_to_theta(start[0], start[1])
        se = self.chr_site_to_theta(start[0], start[2])
        stop = bottom

        es = self.chr_site_to_theta(end[0], end[1])
        ee = self.chr_site_to_theta(end[0], end[2])
        etop = bottom

        z1 = stop - stop * math.cos(abs((se-ss) * 0.5))
        z2 = etop - etop * math.cos(abs((ee-es) * 0.5))

        Path = mpath.Path

        path_data = [(Path.MOVETO,  (ss, bottom)),
                     (Path.CURVE3,  (ss, link_plot_para['center'])),
                     (Path.CURVE3,  (ee,   etop)),
                     (Path.CURVE3,  ((es+ee)*0.5, etop+z2)),
                     (Path.CURVE3,  (es, etop)),
                     (Path.CURVE3,  (es, link_plot_para['center'])),
                     (Path.CURVE3,  (se,   bottom)),
                     (Path.CURVE3,  ((ss+se)*0.5, bottom+z1)),
                     (Path.CURVE3,  (ss, bottom)),
                     ]
        codes, verts = list(zip(*path_data))
        path = mpath.Path(verts, codes)
        patch = mpatches.PathPatch(
            path, facecolor=link_plot_para['facecolor'], alpha=link_plot_para['alpha'], linewidth=link_plot_para['linewidth'], zorder=0)
        self.ax.add_patch(patch)

    def synteny_plot(self, data, bottom, **kwargs):
        """
        data = [("Chr1", 1, 1000, "Chr2", 2000, 3000), ("Chr4", 1, 1000, "Chr7", 3000, 2000)]
        """

        synteny_plot_para = {
            "bottom": bottom,
            "facecolor": None,
            "cmap": None,
            "liner_camp": False,
            "alpha": 0.5,
            "center": 0,
        }

        for k in kwargs:
            if k in synteny_plot_para:
                synteny_plot_para[k] = kwargs[k]

        data = dict(enumerate(data))

        chr_list = list(self.chr_dict.keys())

        chr_sb_dict = {}
        used_id = []
        for chr_id in chr_list:
            chr_sb_dict[chr_id] = []

            for i in data:
                q_c, q_s, q_e, s_c, s_s, s_e = data[i]
                if (q_c == chr_id or s_c == chr_id) and i not in set(used_id):
                    chr_sb_dict[chr_id].append(i)
                    used_id.append(i)

        # plot
        if synteny_plot_para["cmap"]:

            cmap = synteny_plot_para["cmap"]
            color_num = len(
                [i for i in chr_sb_dict if len(chr_sb_dict[i]) > 0])
            color_list = np.linspace(0.0, 1.0, color_num, endpoint=True)

            num = -1
            for chr_id in chr_list:
                if len(chr_sb_dict[chr_id]) > 0:
                    num += 1
                    for i in chr_sb_dict[chr_id]:
                        j = data[i]
                        q_c, s_c = j[0], j[3]
                        if q_c not in self.chr_dict or s_c not in self.chr_dict:
                            continue

                        if not synteny_plot_para['liner_camp']:
                            if hasattr(cmap, 'colors'):
                                cmap_num = len(cmap.colors)
                                self.link_plot(j[0:3], j[3:6], bottom, facecolor=cmap(
                                    (num) % cmap_num), alpha=synteny_plot_para['alpha'])
                            else:
                                self.link_plot(j[0:3], j[3:6], bottom, facecolor=cmap(
                                    num), alpha=synteny_plot_para['alpha'])
                        else:
                            self.link_plot(j[0:3], j[3:6], bottom, facecolor=cmap(
                                color_list[num]), alpha=synteny_plot_para['alpha'])

        elif synteny_plot_para["facecolor"]:
            for i in data:
                j = data[i]
                q_c, s_c = j[0], j[3]
                if q_c not in self.chr_dict or s_c not in self.chr_dict:
                    continue
                self.link_plot(
                    j[0:3], j[3:6], bottom, facecolor=synteny_plot_para["facecolor"], alpha=synteny_plot_para['alpha'])

    def scatter_plot(self, data, bin_length, bottom, height, **kwargs):
        """
        y_axis is a function
        such as:
        y_axis = lambda x:"%.0f%%" % (x * 100)
        """

        scatter_plot_para = {
            "min_plot": None,
            "max_plot": None,
            "linewidth": 0.5,
            "edgecolor": "k",
            "facecolor": "#5b7281",
            "centripetal": False,
            "markersize": 1,
            "y_axis": None,
            "axis_linewidth": 1,
            "axis_color": "k",
            "axis_ticker_num": 5,
            "ticker_length": (2*math.pi/365)*0.5,
            "ticker_label_adjust": 5
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in scatter_plot_para:
                scatter_plot_para[k] = kwargs[k]

        bar_width = bin_length * self.the_vs_len

        num_list = []
        for i in data:
            for j in data[i]:
                num_list.append(data[i][j])

        if scatter_plot_para["max_plot"]:
            max_value = scatter_plot_para["max_plot"]
        else:
            max_value = max(num_list)

        if scatter_plot_para["min_plot"]:
            min_value = scatter_plot_para["min_plot"]
        else:
            min_value = min(num_list)

        he_vs_val = height / abs(max_value - min_value)

        for chr_id in self.chr_dict:
            theta = self.chr_dict[chr_id]['theta']
            chr_length = self.chr_dict[chr_id]['length']

            position_list = []
            value_list = []

            theta_start = theta[0]
            for i, s, e in split_sequence_to_bins(chr_length, bin_length, 1):
                position_list.append(theta_start)

                if i in data[chr_id]:
                    value = data[chr_id][i]
                else:
                    value = min_value

                if value >= max_value:
                    value = max_value

                elif value <= min_value:
                    value = min_value

                if scatter_plot_para["centripetal"]:
                    plot_value = bottom - (value - min_value) * he_vs_val
                else:
                    plot_value = bottom + (value - min_value) * he_vs_val

                value_list.append(plot_value)

                theta_start += bar_width

            self.ax.scatter(position_list, value_list, s=scatter_plot_para["markersize"], c=scatter_plot_para[
                            "facecolor"], linewidth=scatter_plot_para["linewidth"], edgecolor=scatter_plot_para["edgecolor"])

        # add y-axis
        if scatter_plot_para["y_axis"]:
            y_axis_theta = self.chr_dict[list(
                self.chr_dict.keys())[0]]['theta'][0] - (2*math.pi/365)*0.5

            # add spine
            spine_start = bottom
            if scatter_plot_para["centripetal"]:
                spine_end = bottom - height
            else:
                spine_end = bottom + height

            self.ax.plot([y_axis_theta, y_axis_theta], [
                         spine_start, spine_end], color="k", linewidth=scatter_plot_para["axis_linewidth"])

            # add tick
            tick_height_list = [spine_start, spine_end]
            for i in range(scatter_plot_para["axis_ticker_num"]):
                tick_height_list.append(
                    spine_start + i * (spine_end - spine_start) / scatter_plot_para["axis_ticker_num"])

            for tick_theta in tick_height_list:
                self.ax.plot([y_axis_theta, y_axis_theta - scatter_plot_para["ticker_length"]], [
                    tick_theta, tick_theta], color="k", linewidth=scatter_plot_para["axis_linewidth"])

            # add tick label
            tick_label_theta = y_axis_theta - \
                scatter_plot_para["ticker_label_adjust"] * \
                scatter_plot_para["ticker_length"]
            for tick_label_height, tick_label_value in [(spine_start, min_value), (spine_end, max_value)]:

                lab = self.ax.text(tick_label_theta, tick_label_height, scatter_plot_para["y_axis"](tick_label_value), ha='center',
                                   va='center', rotation=np.rad2deg(2*math.pi - (tick_label_theta - 0.5*math.pi)))

    def add_box(self, chr_loci_list, bottom, height, **kwargs):
        # parameter
        # default parameter

        box_parameter = {
            "bottom": bottom,
            "height": height,
            "facecolor": "none",
            "edgecolor": "#000000",
            "linewidth": 1.0,
        }

        # overwrite by given parameter
        for k in kwargs:
            if k in box_parameter:
                box_parameter[k] = kwargs[k]

        # add chr box
        for chr_loci in chr_loci_list:

            pos = self.chr_site_to_theta(chr_loci.chr_id, chr_loci.start)
            width = (chr_loci.end - chr_loci.start) * self.the_vs_len

            height = box_parameter["height"]
            bottom = box_parameter["bottom"]
            facecolor = box_parameter["facecolor"]
            edgecolor = box_parameter["edgecolor"]
            linewidth = box_parameter["linewidth"]

            self.ax.bar([pos], [height], bottom=bottom, width=width, facecolor=facecolor,
                        linewidth=linewidth, edgecolor=edgecolor, align="edge")

    def save(self, file_name="test", format="pdf", dpi=None):
        self.figure.patch.set_alpha(0.0)
        if format == "pdf":
            self.figure.savefig(file_name + ".pdf", bbox_inches="tight")
        elif format == "svg":
            self.figure.savefig(file_name + ".svg", format='svg', facecolor='none', edgecolor='none')
        else:
            if dpi is None:
                dpi = 600
            self.figure.savefig(file_name + "." + format,
                                bbox_inches="tight", dpi=dpi)
        return self.figure


def get_gene_density(gff_file, top_gene_feature, bin_length=1000000, print_stat=False):
    gff_dict = read_gff_file(gff_file)[top_gene_feature]

    chr_list = list(set([gff_dict[i].chr_id for i in gff_dict]))

    chr_interlap_dict = {}
    for chr_id in chr_list:
        chr_interlap_dict[chr_id] = InterLap()
        chr_interlap_dict[chr_id].update(
            [(gff_dict[i].start, gff_dict[i].end) for i in gff_dict if gff_dict[i].chr_id == chr_id])

    gene_density_dict = {}
    for chr_id in chr_interlap_dict:
        gene_density_dict[chr_id] = {}
        chr_len = max(chr_interlap_dict[chr_id])[1]
        for i, s, e in split_sequence_to_bins(chr_len, bin_length, 1):
            gene_density_dict[chr_id][i] = len(
                list(chr_interlap_dict[chr_id].find((s, e))))

    if print_stat:
        num_list = []
        for i in gene_density_dict:
            for j in gene_density_dict[i]:
                num_list.append(gene_density_dict[i][j])

        p_range_list = list(range(0, 110, 10))
        percent_list = np.percentile(np.array(num_list), p_range_list)
        for i in range(len(p_range_list)):
            print("P%d: %.2f" % (p_range_list[i], percent_list[i]))

    return gene_density_dict


def get_cds_density(gff_file, TE_gff_file, genome_fasta_file, bin_length=1000000, print_stat=False):
    fa_dict = read_fasta_by_faidx(genome_fasta_file)
    chr_len_dict = {i: fa_dict[i].len() for i in fa_dict}

    # gene gff
    gff_dict = read_gff_file(gff_file)

    chr_dict = {}
    for i in gff_dict:
        for gene in gff_dict[i]:
            gene = gff_dict[i][gene]
            if gene.chr_id not in chr_dict:
                chr_dict[gene.chr_id] = []
            for j in gene.sub_features:
                for k in j.sub_features:
                    if k.type == 'CDS':
                        chr_dict[gene.chr_id].append((k.start, k.end))

    for i in chr_dict:
        chr_dict[i] = merge_intervals(chr_dict[i])

    chr_interlap_dict = {}
    for chr_id in chr_dict:
        chr_interlap_dict[chr_id] = InterLap()
        chr_interlap_dict[chr_id].update(chr_dict[chr_id])

    # TE gff
    te_gff_dict = read_gff_file(TE_gff_file)

    te_chr_dict = {}
    for i in te_gff_dict:
        for gene in te_gff_dict[i]:
            gene = te_gff_dict[i][gene]
            if gene.chr_id not in te_chr_dict:
                te_chr_dict[gene.chr_id] = []
            te_chr_dict[gene.chr_id].append((gene.start, gene.end))

    for i in te_chr_dict:
        te_chr_dict[i] = merge_intervals(te_chr_dict[i])

    te_chr_interlap_dict = {}
    for chr_id in te_chr_dict:
        te_chr_interlap_dict[chr_id] = InterLap()
        te_chr_interlap_dict[chr_id].update(te_chr_dict[chr_id])

    # remove overlap with te
    from lib.common.math.interval import section

    cds_density_dict = {}
    for chr_id in chr_interlap_dict:
        cds_density_dict[chr_id] = {}
        chr_len = chr_len_dict[chr_id]
        for i, s, e in split_sequence_to_bins(chr_len, bin_length, 1):
            cds_list = list(chr_interlap_dict[chr_id].find((s, e)))
            raw_sum_length = sum_interval_length(cds_list)

            overlap_with_te = []
            for cds in cds_list:
                for te in te_chr_interlap_dict[chr_id].find((cds[0], cds[1])):
                    if_flag, deta = section(cds, te)
                    overlap_with_te.append(deta)

            overlap_with_te_length = sum_interval_length(
                merge_intervals(overlap_with_te))

            cds_density_dict[chr_id][i] = (
                raw_sum_length - overlap_with_te_length) / (e-s+1)

    if print_stat:
        num_list = []
        for i in cds_density_dict:
            for j in cds_density_dict[i]:
                num_list.append(cds_density_dict[i][j])

        p_range_list = list(range(0, 110, 10))
        percent_list = np.percentile(np.array(num_list), p_range_list)
        for i in range(len(p_range_list)):
            print("P%d: %.2f" % (p_range_list[i], percent_list[i]))

    return cds_density_dict


def get_genome_CG_ratio(genome_fasta, bin_length=1000000, print_stat=False):
    fasta_dict = read_fasta_by_faidx(genome_fasta)

    genome_CG_ratio_dict = {}
    for chr_id in fasta_dict:
        genome_CG_ratio_dict[chr_id] = {}
        chr_len = fasta_dict[chr_id].len()
        for i, s, e in split_sequence_to_bins(chr_len, bin_length, 1):
            seq_tmp = fasta_dict[chr_id].sub(s, e, "+", False)
            c = seq_tmp.count('C')
            g = seq_tmp.count('G')
            c_ratio = (c + g) / (e-s+1)
            genome_CG_ratio_dict[chr_id][i] = c_ratio

    if print_stat:
        num_list = []
        for i in genome_CG_ratio_dict:
            for j in genome_CG_ratio_dict[i]:
                num_list.append(genome_CG_ratio_dict[i][j])

        p_range_list = list(range(0, 110, 10))
        percent_list = np.percentile(np.array(num_list), p_range_list)
        for i in range(len(p_range_list)):
            print("P%d: %.2f" % (p_range_list[i], percent_list[i]))

    return genome_CG_ratio_dict


if __name__ == "__main__":
    cp = CirclePlot()
    # name, length, bottom (0<=bottom<=1000), height (0<=bottom<=1000)
    cp.load_chr_info("1", 1000, bottom=500, height=100)
    cp.load_chr_info("2", 2000, bottom=500, height=100, facecolor="#ED665D")
    cp.load_chr_info("3", 3000, bottom=500, height=100, facecolor="#6DCCDA")
    cp.load_chr_info("4", 2000, bottom=400, height=100, facecolor="#ED97CA")
    cp.load_chr_info("5", 5000, bottom=550, height=50, facecolor="#EDC948")
    




import os
import re
from Bio import SeqIO


repbase_browse_url = "https://www.girinst.org/repbase/update/browse.php"









RepBase_merged_fasta_id_file = "/lustre/home/xuyuxing/Database/RepBase/RepBase26.02.merged.fasta.id"
RepBase_embl_dir = "/lustre/home/xuyuxing/Database/RepBase/RepBase26.02.embl"


def embl_parser(RepBase_embl_dir):
    ref_file_list = []
    file_dir_list = os.listdir(RepBase_embl_dir)
    for i in file_dir_list:
        if re.match(r'^.*\.ref$', i):
            file_name = RepBase_embl_dir + "/" + i
            ref_file_list.append(file_name)

    RepBase_appendix_embl_dir = RepBase_embl_dir + "/appendix"
    file_dir_list = os.listdir(RepBase_appendix_embl_dir)
    for i in file_dir_list:
        if re.match(r'^.*\.ref$', i):
            file_name = RepBase_appendix_embl_dir + "/" + i
            ref_file_list.append(file_name)

    ref_file_list_recode = []
    for file_name in ref_file_list:
        with open(file_name+".recode", 'w') as fo:
            with open(file_name, 'r', encoding='cp1252') as fi:
                for each_line in fi:
                    fo.write(each_line)
        ref_file_list_recode.append(file_name+".recode")


    for file_name in ref_file_list_recode:
        for record in SeqIO.parse(file_name, "embl"):
            print(record.name)











with open(RepBase_merged_fasta_id_file, 'r') as f:
    id_info_dict = {}
    group_list = []
    for each_line in f:
        each_line = each_line.strip()

        blast_rec_id = each_line.split()[0]
        
        # normal
        matchobj = re.match(r'^(.*)\t(.*)\t(.*)$', each_line)
        if matchobj:
            rec_id, group_id, sp_id = matchobj.groups()
            group_list.append(group_id)

            if blast_rec_id not in id_info_dict:
                id_info_dict[blast_rec_id] = []
            id_info_dict[blast_rec_id].append(group_id)
            id_info_dict[blast_rec_id] = list(set(id_info_dict[blast_rec_id]))
    group_list = list(set(group_list))
            




from lib.common.os import mkdir, rmdir, cmd_run
import os
import re
from lib.common.genome.seq_base import read_fasta_by_faidx
import uuid


def running_samtools_and_bcftools_for_assembly_evaluator(assembly_fasta_file, ngs_r1, ngs_r2, threads, tmp_dir):
    cmd_run("hisat2-build %s %s.hisat2" % (assembly_fasta_file,
                                           assembly_fasta_file), cwd=tmp_dir, retry_max=1, silence=True)
    hisat_out = cmd_run("hisat2 -p %d -x %s.hisat2 -1 %s -2 %s -S ngs.sam" %
                        (threads, assembly_fasta_file, ngs_r1, ngs_r2), cwd=tmp_dir, retry_max=1, silence=True)
    # hisat_out = hisat_out[2]
    # hisat_map_ratio = float(hisat_out.split("\n")[-2].split(" ")[0].split("%")[0])/100

    cmd_run("samtools view -bS -@ %d -o ngs.bam ngs.sam" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools fixmate -@ %d -m ngs.bam ngs.fixmate.bam" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools sort -@ %d -o ngs.fixmate.sorted.bam ngs.fixmate.bam" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools markdup -@ %d -r ngs.fixmate.sorted.bam ngs.fixmate.sorted.undup.bam" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools index -@ %d ngs.fixmate.sorted.undup.bam" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools stats -@ %d -c 3,200,50 ngs.fixmate.sorted.undup.bam > ngs.fixmate.sorted.undup.stat" %
            threads, cwd=tmp_dir, retry_max=1, silence=True)

    cmd_run("samtools faidx %s" % assembly_fasta_file,
            cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("samtools mpileup -g -f %s ngs.fixmate.sorted.undup.bam > ngs.bcf" %
            assembly_fasta_file, cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("bcftools call -mv -Ov ngs.bcf > ngs.var.bcf",
            cwd=tmp_dir, retry_max=1, silence=True)
    cmd_run("bcftools view ngs.var.bcf | vcfutils.pl varFilter - > ngs.var.final.vcf",
            cwd=tmp_dir, retry_max=1, silence=True)


def assembly_evaluator_main(target_assembly_fasta, given_short_reads_fastq1, given_short_reads_fastq2, output_file, threads=56, tmp_dir=None):
    """
    target_assembly_fasta = "/lustre/home/xuyuxing/Database/Lindenbergia/genome/nextpolish2/01_rundir/00.lgs_polish/input.genome.fasta"
    given_short_reads_fastq1 = "/lustre/home/xuyuxing/Database/Lindenbergia/genome/tmp.1.fq.gz"
    given_short_reads_fastq2 = "/lustre/home/xuyuxing/Database/Lindenbergia/genome/tmp.2.fq.gz"
    threads = 56
    tmp_dir = "/lustre/home/xuyuxing/Database/Lindenbergia/genome/tmp"
    """

    if not tmp_dir:
        tmp_dir = "/tmp/%s" % uuid.uuid1().hex
    else:
        tmp_dir = tmp_dir + "/" + uuid.uuid1().hex
        tmp_dir = os.path.abspath(tmp_dir)

    mkdir(tmp_dir, False)
    assembly_fasta_file = tmp_dir + "/assembly.fasta"
    ngs_r1 = tmp_dir + "/ngs_r1.fq.gz"
    ngs_r2 = tmp_dir + "/ngs_r2.fq.gz"

    os.symlink(os.path.abspath(target_assembly_fasta), assembly_fasta_file)
    os.symlink(os.path.abspath(given_short_reads_fastq1), ngs_r1)
    os.symlink(os.path.abspath(given_short_reads_fastq2), ngs_r2)

    # sum_length
    assembly_seq_dict = read_fasta_by_faidx(tmp_dir + "/assembly.fasta")
    sum_length = sum([assembly_seq_dict[i].len() for i in assembly_seq_dict])

    running_samtools_and_bcftools_for_assembly_evaluator(
        assembly_fasta_file, ngs_r1, ngs_r2, threads, tmp_dir)

    # reads, mapped_reads, paired_mapped_reads, base_error_rate, bases_mapped, uncovered_base
    with open(tmp_dir+"/ngs.fixmate.sorted.undup.stat", 'r') as f:
        covered_base = 0
        for each_line in f:
            mObj = re.match('^SN\treads paired:\t(\S+)', each_line)
            if mObj:
                reads = int(mObj.groups()[0])

            mObj = re.match('^SN\treads mapped:\t(\S+)', each_line)
            if mObj:
                mapped_reads = int(mObj.groups()[0])

            mObj = re.match('^SN\treads mapped and paired:\t(\S+)', each_line)
            if mObj:
                paired_mapped_reads = int(mObj.groups()[0])

            mObj = re.match('^SN\terror rate:\t(\S+)', each_line)
            if mObj:
                base_error_rate = float(mObj.groups()[0])

            mObj = re.match('^COV\t\[\d+-\d+\]\t(\S+)\t(\S+)', each_line)
            if mObj:
                covered_base += int(mObj.groups()[1])

        print(reads, mapped_reads, paired_mapped_reads,
              base_error_rate, covered_base)

    # snp
    hete = int(cmd_run("grep \"^#\" -P ngs.var.final.vcf -v |grep \"1/1\" -c",
                       cwd=tmp_dir, retry_max=1, silence=True)[1].strip())
    error = int(cmd_run("grep \"^#\" -P ngs.var.bcf -v |grep \"1/1\" -c -v",
                        cwd=tmp_dir, retry_max=1, silence=True)[1].strip())

    #
    with open(output_file, 'w') as f:
        f.write("""
#### Assembly Evaluator Report:

Assembly_sum_length:\t\t\t%d
NGS_reads_number:\t\t\t%d
Total_mapped_reads_number:\t\t%d (%.2f%%)
Paired_mapped_reads_number:\t\t%d (%.2f%%)
Base_error_rate:\t\t\t%.5f%%\t\t\t\t\t\t# error rate from samtools stat, for reads error include sequence error
Assembly_covered_by_reads (bp):\t\t%d (%.2f%%)\t\t\t\t\t# at least more than 3 depth
Heterozygosity:\t\t\t\t%d (%.5f%%)\t\t\t\t\t# strict 1/1 SNP, Indel vs assembly covered length
Pure_error_bases:\t\t\t%d (%.5f%%)\t\t\t\t\t\t# no strict 0/1 SNP, Indel vs assembly covered length
""" % (
            sum_length,
            reads,
            mapped_reads, mapped_reads/reads * 100,
            paired_mapped_reads, paired_mapped_reads/reads * 100,
            base_error_rate * 100,
            covered_base, covered_base/sum_length * 100,
            hete, hete/covered_base * 100,
            error, error/covered_base * 100,
        ))

    # remove
#     rmdir(tmp_dir)



from config import plant_tribes_scaffold_dir, plant_tribes_default_scaffold, hmmsearch_path, cap3_path, transdecoder_path, gt_path, trimal_path, cd_hit_dir_path, diamond_path, hmmpress_path, hmmbuild_path, clustalo_path
from lib.common.evolution.orthotools2 import read_species_info, OrthoGroups
from lib.common.fileIO import read_list_file, tsv_file_dict_parse
from lib.common.genome.seq_base import BioSeq, read_fasta_by_faidx, reverse_complement, read_fasta
from lib.common.os import cmd_run, mkdir, multiprocess_running, rmdir, move_file, get_file_name, get_file_dir
from lib.xuyuxing.base.base_function import merge_file
from lib.xuyuxing.base.common_command import merge_dict
from retry import retry
from scipy import stats
import numpy as np
import os
import re

# build scafflod dir


def run_clustalo(in_file, aln_file, cpu_num=1):
    in_file = os.path.abspath(in_file)
    aln_file = os.path.abspath(aln_file)
    work_dir = get_file_dir(aln_file)

    if not(os.path.exists(aln_file) and os.path.getsize(aln_file) != 0):
        cmd_string = clustalo_path + " -i %s -o %s -t Protein --force --auto --threads=%d" % (
            in_file, aln_file, cpu_num)
        cmd_run(cmd_string, cwd=work_dir, silence=True)


def run_hmmbuild(in_file, hmm_file):
    in_file = os.path.abspath(in_file)
    hmm_file = os.path.abspath(hmm_file)
    work_dir = get_file_dir(hmm_file)

    if not(os.path.exists(hmm_file) and os.path.getsize(hmm_file) != 0):
        cmd_string = hmmbuild_path + " %s %s" % (hmm_file, in_file)
        cmd_run(cmd_string, cwd=work_dir, silence=True)


def build_scaffold_dir(orthogroups, species_info_dict, output_dir, num_threads):
    output_dir = os.path.abspath(output_dir)

    mkdir(output_dir, True)

    orthogroups.write_OG_tsv_file(output_dir+"/orthogroups.tsv")

    sub_dir_dict = {
        'alns': output_dir + "/alns",
        'db': output_dir + "/db",
        'fasta': output_dir + "/fasta",
        'hmms': output_dir + "/hmms",
    }

    for i in sub_dir_dict:
        mkdir(sub_dir_dict[i], True)

    sorted_og_list = sorted(orthogroups.OG_id_list, key=lambda x: len(
        orthogroups.OG_dict[x].gene_list), reverse=True)

    # write fasta
    input_aa_fasta_dict = {}
    for sp_id in species_info_dict:
        input_aa_fasta_dict[sp_id] = read_fasta_by_faidx(
            species_info_dict[sp_id].pt_file)

    input_cds_fasta_dict = {}
    for sp_id in species_info_dict:
        if species_info_dict[sp_id].cds_file:
            input_cds_fasta_dict[sp_id] = read_fasta_by_faidx(
                species_info_dict[sp_id].cds_file)

    seq_dir = sub_dir_dict['fasta']
    mkdir(seq_dir, True)

    for raw_og_id in sorted_og_list:
        cds_file = seq_dir + "/%s.fna" % raw_og_id
        aa_file = seq_dir + "/%s.faa" % raw_og_id

        if not(os.path.exists(aa_file) and os.path.getsize(aa_file) != 0):

            cds_f = open(cds_file, 'w')
            aa_f = open(aa_file, 'w')

            og = orthogroups.OG_dict[raw_og_id]
            for sp_id in og.gene_dict:
                gene_list = og.gene_dict[sp_id]
                for gene in gene_list:
                    if sp_id in input_cds_fasta_dict:
                        cds_f.write(">%s\n%s\n" %
                                    (gene.id, input_cds_fasta_dict[sp_id][gene.id].seq))
                    aa_f.write(">%s\n%s\n" %
                               (gene.id, input_aa_fasta_dict[sp_id][gene.id].seq))

            cds_f.close()
            aa_f.close()

    # get aln
    aln_dir = sub_dir_dict['alns']
    mkdir(aln_dir, True)

    args_list = []
    for raw_og_id in sorted_og_list:
        aa_file = seq_dir + "/%s.faa" % raw_og_id
        aln_file = aln_dir + "/%s.aln" % raw_og_id

        args_list.append((aa_file, aln_file))

    multiprocess_running(run_clustalo, args_list, num_threads, silence=True)

    # get hmm
    hmm_dir = sub_dir_dict['hmms']
    mkdir(hmm_dir, True)

    args_list = []
    for raw_og_id in sorted_og_list:
        aln_file = aln_dir + "/%s.aln" % raw_og_id
        hmm_file = hmm_dir + "/%s.hmm" % raw_og_id

        args_list.append((aln_file, hmm_file))

    multiprocess_running(run_hmmbuild, args_list, num_threads, silence=True)

    # get db
    # hmm
    hmm_file_list = [hmm_dir + "/%s.hmm" % raw_og_id
                     for raw_og_id in sorted_og_list if os.path.exists(hmm_dir + "/%s.hmm" % raw_og_id)]

    hmm_db_dir = sub_dir_dict['db'] + "/hmm"
    mkdir(hmm_db_dir, True)

    hmm_db_file = hmm_db_dir + "/hmm_db.hmm"
    merge_file(hmm_file_list, hmm_db_file)

    cmd_string = hmmpress_path + " " + hmm_db_file
    cmd_run(cmd_string, cwd=hmm_db_dir, silence=True)

    # diamond
    aa_file_list = [seq_dir + "/%s.faa" % raw_og_id
                    for raw_og_id in sorted_og_list if os.path.exists(seq_dir + "/%s.faa" % raw_og_id)]

    diamond_db_dir = sub_dir_dict['db'] + "/diamond"
    mkdir(diamond_db_dir, True)

    diamond_db_file = diamond_db_dir + "/diamond_db.faa"
    merge_file(aa_file_list, diamond_db_file)

    cmd_string = "%s makedb --in %s --db %s" % (
        diamond_path, diamond_db_file, diamond_db_file)
    cmd_run(cmd_string, cwd=diamond_db_dir, silence=True)


def mulit_build_scaffold_dir_main(args):
    # read orthogroups
    orthogroups = OrthoGroups(OG_tsv_file=args.orthogroups_file)

    og_list = []
    for og_id in orthogroups.OG_id_list:
        og = orthogroups.OG_dict[og_id]
        g_num = len(og.gene_list)
        s_num = len([i for i in og.species_stat if og.species_stat[i] != 0])
        if g_num >= args.min_gene_number and s_num >= args.min_species_number:
            og_list.append(og)

    orthogroups = OrthoGroups(from_OG_list=og_list)

    # read file of files
    species_info_dict = read_species_info(args.species_info_file)

    # build scaffold dir
    build_scaffold_dir(orthogroups, species_info_dict,
                       args.scaffold_dir, args.num_threads)


# seq_classify
def read_hmm_out_get_subject_id_list(hmm_out_file):
    subject_id_list = []
    with open(hmm_out_file, 'r') as f:
        for each_line in f:
            if not re.match(r'^#', each_line):
                s_id = each_line.split()[0]
                subject_id_list.append(s_id)
    return list(set(subject_id_list))


def read_hmm_out_get_subject_id_score_dict(hmm_out_file):
    score_dict = {}
    with open(hmm_out_file, 'r') as f:
        for each_line in f:
            if not re.match(r'^#', each_line):
                s_id = each_line.split()[0]
                score = float(each_line.split()[5])
                if s_id in score_dict:
                    if score > score_dict[s_id]:
                        score_dict[s_id] = score
                else:
                    score_dict[s_id] = score
    return score_dict


def read_hmm_out_get_query_id_score_dict(hmm_out_file):
    score_dict = {}
    with open(hmm_out_file, 'r') as f:
        for each_line in f:
            if not re.match(r'^#', each_line):
                q_id = each_line.split()[2]
                score = float(each_line.split()[5])
                if q_id in score_dict:
                    if score > score_dict[q_id]:
                        score_dict[q_id] = score
                else:
                    score_dict[q_id] = score
    return score_dict


def seq_in_ortho_stat(protein_seq, scaffold_hmm_file, scaffold_seq_fasta, scaffold_msa_file, work_dir, num_threads=1):
    mkdir(work_dir, True)

    input_fasta = work_dir + "/input.faa"

    with open(input_fasta, 'w') as f:
        f.write(">%s\n%s" % ("input_now", protein_seq.seq))

    # get stats
    # hmm score
    hmm_out_file = work_dir + "/pep.hmm.tbl"
    hmm_log_file = work_dir + "/pep.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, hmm_out_file, hmm_log_file, scaffold_hmm_file, input_fasta)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_hmm_out_file = work_dir + "/backbone.hmm.tbl"
    backbone_hmm_log_file = work_dir + "/backbone.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, backbone_hmm_out_file, backbone_hmm_log_file, scaffold_hmm_file, scaffold_seq_fasta)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_score_dict = read_hmm_out_get_subject_id_score_dict(
        backbone_hmm_out_file)
    pep_score_dict = read_hmm_out_get_subject_id_score_dict(
        hmm_out_file)

    backbone_score = [backbone_score_dict[i] for i in backbone_score_dict]

    avg_backbone_score = sum(backbone_score)/len(backbone_score)
    std_backbone_score = np.std(np.array(backbone_score))

    # aln
    pep_aln_file = work_dir + "/pep.aln"
    cmd_string = "mafft --anysymbol --add %s %s > %s" % (
        input_fasta, scaffold_msa_file, pep_aln_file)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    pep_trim_aln_file = work_dir + "/pep.trim.aln"
    cmd_string = "%s -in %s -out %s -gt 0.1 >/dev/null" % (
        trimal_path, pep_aln_file, pep_trim_aln_file)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_seq_id_list = list(read_fasta_by_faidx(scaffold_seq_fasta).keys())
    input_seq_id_list = list(read_fasta_by_faidx(input_fasta).keys())

    aln_seq_dict = read_fasta_by_faidx(pep_trim_aln_file)

    if not (aln_seq_dict[list(aln_seq_dict.keys())[0]].len() == 0):
        stat_dict = {}
        for seq_id in aln_seq_dict:
            aln_seq = aln_seq_dict[seq_id].seq
            aln_cov = len(aln_seq.replace("-", ""))/len(aln_seq)
            stat_dict[seq_id] = (len(aln_seq.replace("-", "")),
                                 len(aln_seq), aln_cov)

        backbone_len = [stat_dict[i][0]
                        for i in stat_dict if i in backbone_seq_id_list]
        backbone_cov = [stat_dict[i][2]
                        for i in stat_dict if i in backbone_seq_id_list]

        avg_backbone_len = sum(backbone_len)/len(backbone_len)
        std_backbone_len = np.std(np.array(backbone_len))

        avg_backbone_cov = sum(backbone_cov)/len(backbone_cov)
        std_backbone_cov = np.std(np.array(backbone_cov))

    else:
        avg_backbone_len = 0.0
        std_backbone_len = 0.0
        avg_backbone_cov = 0.0
        std_backbone_cov = 0.0
        stat_dict = {i: (0, 0, 0.0) for i in input_seq_id_list}

    i = "input_now"
    if i in pep_score_dict:
        score = pep_score_dict[i]
    else:
        score = 0.0

    # col_list = ["seq_id", "cov", "avg_cov", "sd_cov", "len", "avg_len", "sd_len", "score", "avg_score", "sd_score"]

    rmdir(work_dir)

    return protein_seq.seqname, stat_dict[i][2], avg_backbone_cov, std_backbone_cov, stat_dict[i][0], avg_backbone_len, std_backbone_len, score, avg_backbone_score, std_backbone_score


def make_seqs_stat_file(input_fasta, scaffold_hmm_file, scaffold_seq_fasta, scaffold_msa_file, work_dir, num_threads=1):
    mkdir(work_dir, True)

    # get stats
    # hmm score
    hmm_out_file = work_dir + "/pep.hmm.tbl"
    hmm_log_file = work_dir + "/pep.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, hmm_out_file, hmm_log_file, scaffold_hmm_file, input_fasta)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_hmm_out_file = work_dir + "/backbone.hmm.tbl"
    backbone_hmm_log_file = work_dir + "/backbone.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, backbone_hmm_out_file, backbone_hmm_log_file, scaffold_hmm_file, scaffold_seq_fasta)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_score_dict = read_hmm_out_get_subject_id_score_dict(
        backbone_hmm_out_file)
    pep_score_dict = read_hmm_out_get_subject_id_score_dict(
        hmm_out_file)

    backbone_score = [backbone_score_dict[i] for i in backbone_score_dict]

    avg_backbone_score = sum(backbone_score)/len(backbone_score)
    std_backbone_score = np.std(np.array(backbone_score))

    # aln
    pep_aln_file = work_dir + "/pep.aln"
    cmd_string = "mafft --anysymbol --add %s %s > %s" % (
        input_fasta, scaffold_msa_file, pep_aln_file)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    pep_trim_aln_file = work_dir + "/pep.trim.aln"
    cmd_string = "%s -in %s -out %s -gt 0.1 >/dev/null" % (
        trimal_path, pep_aln_file, pep_trim_aln_file)
    cmd_run(cmd_string, silence=True, cwd=work_dir)

    backbone_seq_id_list = list(read_fasta_by_faidx(scaffold_seq_fasta).keys())
    input_seq_id_list = list(read_fasta_by_faidx(input_fasta).keys())

    aln_seq_dict = read_fasta_by_faidx(pep_trim_aln_file)

    if not (aln_seq_dict[list(aln_seq_dict.keys())[0]].len() == 0):
        stat_dict = {}
        for seq_id in aln_seq_dict:
            aln_seq = aln_seq_dict[seq_id].seq
            aln_cov = len(aln_seq.replace("-", ""))/len(aln_seq)
            stat_dict[seq_id] = (len(aln_seq.replace("-", "")),
                                 len(aln_seq), aln_cov)

        backbone_len = [stat_dict[i][0]
                        for i in stat_dict if i in backbone_seq_id_list]
        backbone_cov = [stat_dict[i][2]
                        for i in stat_dict if i in backbone_seq_id_list]

        avg_backbone_len = sum(backbone_len)/len(backbone_len)
        std_backbone_len = np.std(np.array(backbone_len))

        avg_backbone_cov = sum(backbone_cov)/len(backbone_cov)
        std_backbone_cov = np.std(np.array(backbone_cov))

    else:
        avg_backbone_len = 0.0
        std_backbone_len = 0.0
        avg_backbone_cov = 0.0
        std_backbone_cov = 0.0
        stat_dict = {i: (0, 0, 0.0) for i in input_seq_id_list}

    with open(work_dir+"/stats.txt", 'w') as f:
        col_list = ["seq_id", "cov", "avg_cov",
                    "sd_cov", "len", "avg_len", "sd_len", "score", "avg_score", "sd_score"]
        f.write("\t".join(col_list)+"\n")

        for i in input_seq_id_list:
            if i in pep_score_dict:
                score = pep_score_dict[i]
            else:
                score = 0.0

            write_list = [i, "%.2f" % stat_dict[i][2], "%.2f" % avg_backbone_cov, "%.2f" % std_backbone_cov,
                          "%.2f" % stat_dict[i][0], "%.2f" % avg_backbone_len, "%.2f" % std_backbone_len,
                          "%.2f" % score, "%.2f" % avg_backbone_score, "%.2f" % std_backbone_score, ]
            f.write("\t".join(write_list)+"\n")


def get_pass_seq_id(stats_file, p_value=0.05):
    pass_seq_list = []

    if os.path.exists(stats_file) and os.path.getsize(stats_file) != 0:

        stats_info_dict = tsv_file_dict_parse(stats_file, key_col='seq_id')

        for seq_id in stats_info_dict:
            seq_info = stats_info_dict[seq_id]
            pass_flag = True
            for i in ['cov', 'score']:
                seq_value = float(seq_info[i])
                avg_value = float(seq_info['avg_'+i])
                sd_value = float(seq_info['sd_'+i])

                if avg_value > 0.0:
                    distn = stats.norm(avg_value, sd_value)
                    if seq_value < distn.ppf(p_value):
                        pass_flag = False
                else:
                    pass_flag = False

            if pass_flag:
                pass_seq_list.append(seq_id)

    return pass_seq_list


def if_seq_in_ortho(protein_seq, scaffold_hmm_file, scaffold_seq_fasta, scaffold_msa_file, work_dir, clean_up=True, num_threads=1):
    scaffold_hmm_file = os.path.abspath(scaffold_hmm_file)
    scaffold_seq_fasta = os.path.abspath(scaffold_seq_fasta)
    scaffold_msa_file = os.path.abspath(scaffold_msa_file)
    work_dir = os.path.abspath(work_dir)

    mkdir(work_dir, True)

    input_fasta = work_dir + "/input.faa"

    with open(input_fasta, 'w') as f:
        f.write(">%s\n%s" % ("input_now", protein_seq.seq))

    make_seqs_stat_file(input_fasta, scaffold_hmm_file,
                        scaffold_seq_fasta, scaffold_msa_file, work_dir, num_threads=num_threads)

    stat_file = work_dir + "/stats.txt"

    pass_id_list = get_pass_seq_id(stat_file)

    if clean_up:
        rmdir(work_dir)

    if len(pass_id_list) > 0:
        return True
    else:
        return False


def get_best_ortho(seqname, hmm_results, diamond_out, scaffold_dir):
    hmm_score_dict = read_hmm_out_get_query_id_score_dict(hmm_results)

    if len(hmm_score_dict) > 0:
        hmm_top_og_id = sorted(
            hmm_score_dict, key=lambda x: hmm_score_dict[x], reverse=True)[0]
    else:
        hmm_top_og_id = None

    diamond_top_gene_id = None
    with open(diamond_out, 'r') as f:
        all_info = f.read()
        if len(all_info) > 0:
            diamond_top_gene_id = all_info.split("\t")[1]

    if diamond_top_gene_id:
        og_tsv = scaffold_dir + "/orthogroups.tsv"
        with open(og_tsv, 'r') as f:
            for each_line in f:
                if diamond_top_gene_id in each_line:
                    diamond_top_og_id = each_line.split("\t")[0]
    else:
        diamond_top_og_id = None

    if hmm_top_og_id:
        top_og_id = hmm_top_og_id
    elif hmm_top_og_id is None:
        top_og_id = diamond_top_og_id

    return top_og_id


@retry(tries=5)
def seq_classify(protein_seq, scaffold_dir, work_dir, clean_up=True):
    mkdir(work_dir, True)
    input_fasta = work_dir + "/input.faa"

    with open(input_fasta, 'w') as f:
        f.write(">%s\n%s" % (protein_seq.seqname, protein_seq.seq))

    # get best ortho
    hmm_db = scaffold_dir + "/db/hmm/hmm_db.hmm"

    hmm_results = work_dir + "/hmm.tbl"
    hmm_log = work_dir + "/hmm.log"

    cmd_string = '%s -E 1e-20 --cpu 1 --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, hmm_results, hmm_log, hmm_db, input_fasta)
    cmd_run(cmd_string, cwd=work_dir, silence=True)

    diamond_db = scaffold_dir + "/db/diamond/diamond_db.faa.dmnd"

    diamond_out = work_dir + "/diamond.bls"

    cmd_string = "diamond blastp -q %s -k 50 -e 1e-5 -d %s -o %s -f 6 -p 1" % (
        input_fasta, diamond_db, diamond_out)

    cmd_run(cmd_string, cwd=work_dir, silence=True)

    best_ortho = get_best_ortho(
        protein_seq.seqname, hmm_results, diamond_out, scaffold_dir)

    pass_flag = False
    if best_ortho:
        scaffold_hmm_file = "%s/hmms/%s.hmm" % (scaffold_dir, best_ortho)
        scaffold_seq_fasta = "%s/fasta/%s.faa" % (scaffold_dir, best_ortho)
        scaffold_msa_file = "%s/alns/%s.aln" % (scaffold_dir, best_ortho)
        check_dir = work_dir + "/check_dir"

        pass_flag = if_seq_in_ortho(protein_seq, scaffold_hmm_file,
                                    scaffold_seq_fasta, scaffold_msa_file, check_dir, clean_up, 1)

    if clean_up:
        rmdir(work_dir)

    if best_ortho and pass_flag:
        return best_ortho
    else:
        return None


# deep assembly
def clean_assembly_dir(assembly_dir):
    results_dir_flag = False
    for i in os.listdir(assembly_dir):
        if not i in ['results_dir', "pep.fa", "cDNA.fa", "cds.fa", "stats.txt"]:
            rmdir(assembly_dir+"/"+i)
        if i == 'results_dir':
            results_dir_flag = True

    if results_dir_flag:
        for i in ["pep.fa", "cDNA.fa", "cds.fa", "stats.txt"]:
            if os.path.exists(assembly_dir+"/results_dir/"+i):
                move_file(assembly_dir+"/results_dir/"+i, assembly_dir+"/"+i)
        rmdir(assembly_dir+"/results_dir")


@retry(tries=5)
def one_family_assembly(cdna_file, tran_pep_file, hmm_file, aln_file, seq_file, out_dir, out_reads_prefix, transdecoder_id=False, num_threads=1, clean_up=True):
    """
    hmm_file = '/lustre/home/xuyuxing/Program/PlantTribes/PlantTribes_scaffolds/data/26Gv2.0/hmms/orthofinder/25.hmm'
    aln_file = '/lustre/home/xuyuxing/Program/PlantTribes/PlantTribes_scaffolds/data/26Gv2.0/alns/orthofinder/25.aln'
    seq_file = '/lustre/home/xuyuxing/Program/PlantTribes/PlantTribes_scaffolds/data/26Gv2.0/fasta/orthofinder/25.faa'
    out_dir = '/lustre/home/xuyuxing/Work/Orobanchaceae/other_species_trans/Cistanche_deserticola/AssemblyPostProcessor/assembly_dir/25'
    tran_pep_file = '/lustre/home/xuyuxing/Work/Orobanchaceae/other_species_trans/Cistanche_deserticola/AssemblyPostProcessor/Trinity.fasta.transdecoder_dir/longest_orfs.pep'
    cdna_file = '/lustre/home/xuyuxing/Work/Orobanchaceae/other_species_trans/Cistanche_deserticola/Cistanche_deserticola.trinity2.out3/Trinity.fasta'
    out_reads_prefix = 'planttribes_25_'
    transdecoder_id=True
    num_threads=1
    clean_up=True
    """

    already_flag = True
    for i in ["pep.fa", "cDNA.fa", "cds.fa", "stats.txt"]:
        i = out_dir + "/" + i
        if not (os.path.exists(i) and os.path.getsize(i)):
            already_flag = False

    if already_flag:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    mkdir(out_dir, False)

    # hmm run
    hmm_out_file = out_dir + "/hmm.tbl"
    hmm_log_file = out_dir + "/hmm.log"
    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, hmm_out_file, hmm_log_file, hmm_file, tran_pep_file)
    cmd_run(cmd_string, silence=True)
    subject_id_list = read_hmm_out_get_subject_id_list(hmm_out_file)

    if len(subject_id_list) == 0:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    # cap run
    subject_seq_file = out_dir + "/hmm.subject.fasta"
    seq_dict = read_fasta_by_faidx(cdna_file)

    subject_seq_dict = {}
    if transdecoder_id:
        for i in subject_id_list:
            cdna_id = i.split(".")[0]
            subject_seq_dict[cdna_id] = seq_dict[cdna_id].seq
    else:
        for i in subject_id_list:
            subject_seq_dict[i] = seq_dict[i].seq

    with open(subject_seq_file, 'w') as f:
        for s_id in subject_seq_dict:
            f.write(">%s\n%s\n" % (s_id, subject_seq_dict[s_id]))

    cmd_string = "%s %s -o 40 -p 90 >/dev/null" % (cap3_path, subject_seq_file)
    cmd_run(cmd_string, silence=True)

    cap_contig_file = out_dir + "/hmm.subject.fasta.cap.contigs"
    cap_singlet_file = out_dir + "/hmm.subject.fasta.cap.singlets"

    contig_dict = {}
    if os.path.getsize(cap_contig_file):
        contig_dict = read_fasta_by_faidx(cap_contig_file)

    singlet_dict = {}
    if os.path.getsize(cap_singlet_file):
        singlet_dict = read_fasta_by_faidx(cap_singlet_file)

    assembly_seq_file = out_dir + "/cap3.assembly.fasta"
    assembly_seq_dict = {}
    num = 0
    with open(assembly_seq_file, 'w') as f:
        for s_id in contig_dict:
            num += 1
            new_s_id = out_reads_prefix + str(num)
            f.write(">%s\n%s\n" % (new_s_id, contig_dict[s_id].seq))
            assembly_seq_dict[new_s_id] = contig_dict[s_id].seq
        for s_id in singlet_dict:
            num += 1
            new_s_id = out_reads_prefix + str(num)
            f.write(">%s\n%s\n" % (new_s_id, singlet_dict[s_id].seq))
            assembly_seq_dict[new_s_id] = singlet_dict[s_id].seq

    # transdecoder
    cmd_string = '%s -t %s >/dev/null' % (transdecoder_path, assembly_seq_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)
    longest_orfs_pep_file = out_dir + \
        "/cap3.assembly.fasta.transdecoder_dir/longest_orfs.pep"
    longest_orfs_gff_file = out_dir + \
        "/cap3.assembly.fasta.transdecoder_dir/longest_orfs.gff3"
    longest_orfs_cds_file = out_dir + \
        "/cap3.assembly.fasta.transdecoder_dir/longest_orfs.cds"

    if os.path.getsize(longest_orfs_pep_file) == 0:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    # filter low hmm pep
    filter_hmm_out_file = out_dir + \
        "/cap3.assembly.fasta.transdecoder_dir/longest_orfs.pep.hmm.tbl"
    filter_hmm_log_file = out_dir + \
        "/cap3.assembly.fasta.transdecoder_dir/longest_orfs.pep.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, filter_hmm_out_file, filter_hmm_log_file, hmm_file, longest_orfs_pep_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir +
            "/cap3.assembly.fasta.transdecoder_dir")
    good_subject_id_list = read_hmm_out_get_subject_id_list(
        filter_hmm_out_file)

    if len(good_subject_id_list) == 0:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    # get model (longest) protein sequence
    longest_orfs_pep_dict = read_fasta_by_faidx(longest_orfs_pep_file)
    longest_orfs_cds_dict = read_fasta_by_faidx(longest_orfs_cds_file)

    candi_seq_dict = {}
    for seq_id in good_subject_id_list:
        parent_id, sub_id, strand = re.findall(
            r'(.*)(\.p\d+).*\((\S+)\)$', longest_orfs_pep_dict[seq_id].seqname)[0]
        cdna_seq = assembly_seq_dict[parent_id] if strand == '+' else reverse_complement(
            assembly_seq_dict[parent_id])

        pep_seq = longest_orfs_pep_dict[seq_id].seq
        cds_seq = longest_orfs_cds_dict[seq_id].seq
        if pep_seq[-1] == '*':
            pep_seq = pep_seq[:-1]
            cds_seq = cds_seq[:-3]

        if len(cds_seq)/len(pep_seq) != 3.0:
            continue

        if not parent_id in candi_seq_dict:
            candi_seq_dict[parent_id] = {
                'cdna': cdna_seq,
                'cds': cds_seq,
                'pep': pep_seq,
            }
        elif len(longest_orfs_pep_dict[seq_id].seq) > len(candi_seq_dict[parent_id]['pep']):
            candi_seq_dict[parent_id] = {
                'cdna': cdna_seq,
                'cds': cds_seq,
                'pep': pep_seq,
            }

    if len(candi_seq_dict) == 0:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    # dereplicate
    candi_cds_file = out_dir + "/candi_cds.fa"
    with open(candi_cds_file, 'w') as f:
        for parent_id in candi_seq_dict:
            f.write(">%s\n%s\n" %
                    (parent_id, candi_seq_dict[parent_id]['cds']))

    uniq_cds_file = out_dir + "/uniq_cds.fa"
    cmd_string = "%s sequniq -force -o %s %s >/dev/null" % (
        gt_path, uniq_cds_file, candi_cds_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)
    uniq_id_list = list(read_fasta_by_faidx(uniq_cds_file).keys())

    uniq_pep_file = out_dir + "/uniq_pep.fa"
    with open(uniq_pep_file, 'w') as f:
        for parent_id in uniq_id_list:
            f.write(">%s\n%s\n" %
                    (parent_id, candi_seq_dict[parent_id]['pep']))

    # get stats
    # hmm score
    uniq_pep_hmm_out_file = out_dir + "/uniq_pep.hmm.tbl"
    uniq_pep_hmm_log_file = out_dir + "/uniq_pep.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, uniq_pep_hmm_out_file, uniq_pep_hmm_log_file, hmm_file, uniq_pep_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)

    backbone_hmm_out_file = out_dir + "/backbone.hmm.tbl"
    backbone_hmm_log_file = out_dir + "/backbone.hmm.log"

    cmd_string = '%s -E 1e-20 --cpu %d --noali --tblout %s -o %s %s %s >/dev/null' % (
        hmmsearch_path, num_threads, backbone_hmm_out_file, backbone_hmm_log_file, hmm_file, seq_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)

    backbone_score_dict = read_hmm_out_get_subject_id_score_dict(
        backbone_hmm_out_file)
    uniq_pep_score_dict = read_hmm_out_get_subject_id_score_dict(
        uniq_pep_hmm_out_file)

    backbone_score = [backbone_score_dict[i] for i in backbone_score_dict]

    avg_backbone_score = sum(backbone_score)/len(backbone_score)
    std_backbone_score = np.std(np.array(backbone_score))

    ##
    uniq_pep_aln_file = out_dir + "/uniq_pep.aln"
    cmd_string = "mafft --anysymbol --add %s %s > %s" % (
        uniq_pep_file, aln_file, uniq_pep_aln_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)

    uniq_pep_trim_aln_file = out_dir + "/uniq_pep.trim.aln"
    cmd_string = "%s -in %s -out %s -gt 0.1 >/dev/null" % (
        trimal_path, uniq_pep_aln_file, uniq_pep_trim_aln_file)
    cmd_run(cmd_string, silence=True, cwd=out_dir)

    aln_seq_dict = read_fasta_by_faidx(uniq_pep_trim_aln_file)

    if aln_seq_dict[list(aln_seq_dict.keys())[0]].len() == 0:
        if clean_up:
            clean_assembly_dir(out_dir)
        return None

    stat_dict = {}
    for seq_id in aln_seq_dict:
        aln_seq = aln_seq_dict[seq_id].seq
        aln_cov = len(aln_seq.replace("-", ""))/len(aln_seq)
        stat_dict[seq_id] = (len(aln_seq.replace("-", "")),
                             len(aln_seq), aln_cov)

    backbone_len = [stat_dict[i][0]
                    for i in stat_dict if i not in uniq_id_list]
    backbone_cov = [stat_dict[i][2]
                    for i in stat_dict if i not in uniq_id_list]

    avg_backbone_len = sum(backbone_len)/len(backbone_len)
    std_backbone_len = np.std(np.array(backbone_len))

    avg_backbone_cov = sum(backbone_cov)/len(backbone_cov)
    std_backbone_cov = np.std(np.array(backbone_cov))

    results_dir = out_dir + "/results_dir"
    mkdir(results_dir, True)

    with open(results_dir+"/pep.fa", 'w') as f:
        for i in uniq_id_list:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['pep']))

    with open(results_dir+"/cDNA.fa", 'w') as f:
        for i in uniq_id_list:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['cdna']))

    with open(results_dir+"/cds.fa", 'w') as f:
        for i in uniq_id_list:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['cds']))

    with open(results_dir+"/stats.txt", 'w') as f:
        col_list = ["seq_id", "cov", "avg_cov",
                    "sd_cov", "len", "avg_len", "sd_len", "score", "avg_score", "sd_score"]
        f.write("\t".join(col_list)+"\n")

        uniq_id_list = [i for i in uniq_id_list if i in uniq_pep_score_dict]
        sorted_uniq_id_list = sorted(
            uniq_id_list, key=lambda x: uniq_pep_score_dict[x], reverse=True)

        for i in sorted_uniq_id_list:
            write_list = [i, "%.2f" % stat_dict[i][2], "%.2f" % avg_backbone_cov, "%.2f" % std_backbone_cov,
                          "%.2f" % stat_dict[i][0], "%.2f" % avg_backbone_len, "%.2f" % std_backbone_len,
                          "%.2f" % uniq_pep_score_dict[i], "%.2f" % avg_backbone_score, "%.2f" % std_backbone_score, ]
            f.write("\t".join(write_list)+"\n")

    if clean_up:
        for i in os.listdir(out_dir):
            if i != 'results_dir':
                rmdir(out_dir+"/"+i)

        move_file(results_dir+"/pep.fa", out_dir+"/pep.fa")
        move_file(results_dir+"/cDNA.fa", out_dir+"/cDNA.fa")
        move_file(results_dir+"/cds.fa", out_dir+"/cds.fa")
        move_file(results_dir+"/stats.txt", out_dir+"/stats.txt")

        rmdir(results_dir)


def targeted_gene_family_assembly(args):
    args.output_dir = os.path.abspath(args.output_dir)
    mkdir(args.output_dir, True)

    # copy_input_file
    input_fasta = args.output_dir+"/"+get_file_name(args.input_fasta)
    if not os.path.exists(input_fasta):
        os.symlink(os.path.abspath(args.input_fasta), input_fasta)

    # transdecoder
    longest_orfs_pep_file = input_fasta + \
        ".transdecoder_dir/longest_orfs.pep"
    longest_orfs_cds_file = input_fasta + \
        ".transdecoder_dir/longest_orfs.cds"

    if not (os.path.exists(longest_orfs_pep_file) and os.path.getsize(longest_orfs_pep_file)):
        cmd_string = '%s -t %s >/dev/null' % (transdecoder_path, input_fasta)
        cmd_run(cmd_string, silence=True, cwd=args.output_dir)

    # read plant_tribes_scaffold_dir
    orthofinder_id_file = plant_tribes_scaffold_dir + "/data/" + \
        plant_tribes_default_scaffold + "/annot/orthofinder.id"
    scaffold_hmms_dir = plant_tribes_scaffold_dir + "/data/" + \
        plant_tribes_default_scaffold + "/hmms/orthofinder"
    scaffold_seqs_dir = plant_tribes_scaffold_dir + "/data/" + \
        plant_tribes_default_scaffold + "/fasta/orthofinder"
    scaffold_alns_dir = plant_tribes_scaffold_dir + "/data/" + \
        plant_tribes_default_scaffold + "/alns/orthofinder"

    og_id_list = read_list_file(orthofinder_id_file, ignore_head=1)

    # family_assembly
    family_assembly_dir = args.output_dir + "/assembly_dir"
    mkdir(family_assembly_dir, True)

    args_list = []
    args_id_list = []
    for og_id in og_id_list:
        hmm_file = "%s/%s.hmm" % (scaffold_hmms_dir, og_id)
        aln_file = "%s/%s.aln" % (scaffold_alns_dir, og_id)
        seq_file = "%s/%s.faa" % (scaffold_seqs_dir, og_id)
        out_dir = "%s/%s" % (family_assembly_dir, og_id)
        out_reads_prefix = "planttribes_%s_" % og_id
        args_list.append((input_fasta, longest_orfs_pep_file, hmm_file,
                          aln_file, seq_file, out_dir, out_reads_prefix, True))

    multiprocess_running(one_family_assembly, args_list,
                         args.thread_number, silence=False, args_id_list=args_id_list)

    # get stats
    args_list = []
    args_id_list = []
    for og_id in og_id_list:

        stats_file = "%s/%s/stats.txt" % (family_assembly_dir, og_id)

        args_list.append((stats_file,))
        args_id_list.append(og_id)

    mlt_dict = multiprocess_running(
        get_pass_seq_id, args_list, args.thread_number, args_id_list=args_id_list)

    pass_seq_dict = {}
    for og_id in mlt_dict:
        pass_seq_dict[og_id] = mlt_dict[og_id]['output']

    good_og_num = 0
    bad_og_num = 0
    for i in pass_seq_dict:
        if len(pass_seq_dict[i]) > 0:
            good_og_num += 1
        else:
            bad_og_num += 1

    print("There are %d orthogroups, %d can get good assemblies, %d failed" %
          (good_og_num + bad_og_num, good_og_num, bad_og_num))

    # get Non-redundant sequence
    nr_dict = args.output_dir + "/non_redundant"
    mkdir(nr_dict, True)

    longest_orfs_pep_dict = read_fasta_by_faidx(longest_orfs_pep_file)
    longest_orfs_cds_dict = read_fasta_by_faidx(longest_orfs_cds_file)
    input_seq_dict = read_fasta_by_faidx(input_fasta)

    all_in_one_pep_file = nr_dict + "/all_in_one.faa"

    ass_big_pt_dict = {}
    ass_big_cds_dict = {}
    ass_big_cdna_dict = {}

    with open(all_in_one_pep_file, 'w') as f:
        for seq_id in longest_orfs_pep_dict:
            f.write(">%s\n%s\n" % (seq_id, longest_orfs_pep_dict[seq_id].seq))

        num = 0
        for og_id in og_id_list:
            num += 1
            if num % 1000 == 0:
                print(num)
            pep_file = "%s/%s/pep.fa" % (family_assembly_dir, og_id)
            cdna_file = "%s/%s/cDNA.fa" % (family_assembly_dir, og_id)
            cds_file = "%s/%s/cds.fa" % (family_assembly_dir, og_id)
            if os.path.exists(pep_file):
                pep_dict, seqname_list = read_fasta(pep_file)
                for seq_id in pep_dict:
                    f.write(">%s\n%s\n" % (seq_id, pep_dict[seq_id].seq))

                cdna_dict, seqname_list = read_fasta(cdna_file)
                cds_dict, seqname_list = read_fasta(cds_file)

                ass_big_pt_dict = merge_dict(
                    [ass_big_pt_dict, pep_dict], False)
                ass_big_cds_dict = merge_dict(
                    [ass_big_cds_dict, cds_dict], False)
                ass_big_cdna_dict = merge_dict(
                    [ass_big_cdna_dict, cdna_dict], False)

    cd_hit_out = nr_dict + "/cd_hit_out.100.pep"
    cmd_string = '%s/cd-hit -i %s -o %s -c 1.0 -n 5 -M 16000 -d 0 -T 8 >/dev/null' % (
        cd_hit_dir_path, all_in_one_pep_file, cd_hit_out)
    cmd_run(cmd_string, silence=True, cwd=nr_dict)

    # get output
    cd_hit_seq_dict = read_fasta_by_faidx(cd_hit_out)

    candi_seq_dict = {}
    num = 0
    for seq_id in cd_hit_seq_dict:
        if 'planttribes' in seq_id:
            cds_seq = ass_big_cds_dict[seq_id].seq
            pep_seq = ass_big_pt_dict[seq_id].seq
            cdna_seq = ass_big_cdna_dict[seq_id].seq

            if len(cds_seq)/len(pep_seq) != 3.0:
                continue

        else:
            seq_record = longest_orfs_pep_dict[seq_id]

            parent_id, sub_id, strand = re.findall(
                r'(.*)(\.p\d+).*\((\S+)\)$', seq_record.seqname)[0]
            cdna_seq = input_seq_dict[parent_id].seq if strand == '+' else reverse_complement(
                input_seq_dict[parent_id].seq)

            pep_seq = longest_orfs_pep_dict[seq_id].seq
            cds_seq = longest_orfs_cds_dict[seq_id].seq
            if pep_seq[-1] == '*':
                pep_seq = pep_seq[:-1]
                cds_seq = cds_seq[:-3]

            if len(cds_seq)/len(pep_seq) != 3.0:
                continue

        candi_seq_dict["Uniq_" + str(num)] = {
            'pep': pep_seq,
            'cdna': cdna_seq,
            'cds': cds_seq,
        }

        num += 1

    with open(args.output_dir+"/pep.fa", 'w') as f:
        for i in candi_seq_dict:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['pep']))

    with open(args.output_dir+"/cDNA.fa", 'w') as f:
        for i in candi_seq_dict:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['cdna']))

    with open(args.output_dir+"/cds.fa", 'w') as f:
        for i in candi_seq_dict:
            f.write(">%s\n%s\n" % (i, candi_seq_dict[i]['cds']))


def GetDeepAssemblyStats_main(args):
    orthofinder_id_file = plant_tribes_scaffold_dir + "/data/" + \
        plant_tribes_default_scaffold + "/annot/orthofinder.id"
    family_assembly_dir = args.family_dir

    og_id_list = read_list_file(orthofinder_id_file, ignore_head=1)

    # get stats
    args_list = []
    args_id_list = []
    for og_id in og_id_list:

        stats_file = "%s/%s/stats.txt" % (family_assembly_dir, og_id)

        args_list.append((stats_file,))
        args_id_list.append(og_id)

    mlt_dict = multiprocess_running(
        get_pass_seq_id, args_list, args.thread_number, args_id_list=args_id_list)

    pass_seq_dict = {}
    for og_id in mlt_dict:
        pass_seq_dict[og_id] = mlt_dict[og_id]['output']

    good_og_num = 0
    bad_og_num = 0
    for i in pass_seq_dict:
        if len(pass_seq_dict[i]) > 0:
            good_og_num += 1
        else:
            bad_og_num += 1

    print("There are %d orthogroups, %d can get good assemblies, %d failed" %
          (good_og_num + bad_og_num, good_og_num, bad_og_num))


def MapGenomeToOrthoGroups_main():
    pass


if __name__ == '__main__':
    from lib.common.evolution.orthotools2 import read_species_info, OrthoGroups

    species_info_file = "/lustre/home/xuyuxing/Work/Orobanchaceae/gene_loss/Oro_ref.xlsx"
    orthogroups_file = "/lustre/home/xuyuxing/Work/Orobanchaceae/gene_loss/tree/orthologous_OGs.tsv"
    num_threads = 56

    # read orthogroups
    orthogroups = OrthoGroups(OG_tsv_file=orthogroups_file)

    og_list = []
    for i in orthogroups.OG_id_list[18000:20000]:
        og_list.append(orthogroups.OG_dict[i])

    orthogroups = OrthoGroups(from_OG_list=og_list)

    # read file of files
    species_info_dict = read_species_info(species_info_file)

    # build scaffold dir
    scaffold_dir = "/lustre/home/xuyuxing/Work/Orobanchaceae/gene_loss/gene_loss/Ocu/test2"
    build_scaffold_dir(orthogroups, species_info_dict,
                       scaffold_dir, num_threads)

    # classify a sequence
    from lib.common.genome.seq_base import BioSeq
    protein_seq = BioSeq("MGQCYGKTIPTGDNDGPATTITTIAAAAEDLPEQTPPRNGTPSVKNTPARSSANSPWPSPYPSAAGTPAGVSPSTARSTPRRFFKKPFPPPSPAKHIRASLRKLGQRKKPPREGPIPEDADAGESEQQQQHALDKNFGYNKNFGAKYELGKEVGRGHFGHTCCAKGRKGELKDIPLAVKIISKAQFEKRTGHHECDSDLDMSDVKLIKDHKSEEDFRLLDPNLIDIFRFGYLQSYETGSYLIKAMVFMMTTAISIEDVRREVKILKALSGHRHLASFYDACEDSNNVYIIMELCEGGELLDRILAKGGKYSEDEAKLIIVQILSVVSFCHLQGVVHRDLKPENFLFTSRSEDADLKLIDFGLSDFIRTVGCLSLCIISHVLILSDERLNDIVGSAYYVAPEVLHRSYSVEADIWSIGVIAYILLCGSRPFWARTESGIFRAVLRADPNFEDLPWPSVSLEAKDFVRRLLNKDYRKRMTAAQALTHPWLRSESHPIPLDILVYKLVKSYLHATPFKRAALKALSKALTEDELIYLRAQFMLLEPSEDGRVSLENFRKISGSSHLALSLRNNILLPNTLELCDFQLLFANTAGLLHFSIFQALARNATDALNMSRVPDILNSMAPLSFRRMDLEEFCAAAISTYQLEALENWEQIASTAFEFFEQEGNRAISVEELARELNVGPSAHPMLRDWLRRDGKLSLVGYTKFLHGLTLRSSNMRHH", seqname="T4170N0C00004G00375")
    scaffold_dir = "/lustre/home/xuyuxing/Work/Orobanchaceae/gene_loss/gene_loss/Ocu/test2"
    work_dir = "/lustre/home/xuyuxing/Work/Orobanchaceae/gene_loss/gene_loss/Ocu/test2/test"
    clean_up = True

    # improve trinity assembly
    class abc():
        pass

    args = abc()

    args.input_fasta = "/lustre/home/xuyuxing/Work/Orobanchaceae/HGT/database/1kp/ERS1829209/ERS1829209.fasta"
    args.output_dir = "/lustre/home/xuyuxing/Work/Orobanchaceae/HGT/database/1kp/ERS1829209/AssemblyPostProcessor"
    args.thread_number = 56

    targeted_gene_family_assembly(args)



import re
import lib.common.sqlite_command as sc
import sqlite3
import copy
from Bio.Phylo.BaseTree import Tree, Clade
from lib.common.evolution.tree_operate import remove_pass_node, lookup_by_names, get_offspring, draw_ascii, add_clade_name, \
    get_parent, get_ancestors, get_MRCA, get_sons
from lib.xuyuxing.math.stats import get_qvalue
from scipy import stats
from lib.common.util import printer_list
from mlxtend.evaluate import permutation_test
import numpy as np
from lib.common.math.set import merge_subset
from itertools import combinations


class Taxon(object):
    """
    Taxon is a class for each taxon in taxonomy database.
    """

    def __init__(self):
        pass

    # this list is from https://en.wikipedia.org/wiki/Taxonomic_rank
    rank_list = ["superkingdom", "kingdom", "subkingdom", "superphylum", "phylum", "subphylum", "superclass", "class",
                 "subclass", "infraclass", "cohort", "subcohort", "superorder", "order", "suborder", "infraorder",
                 "parvorder", "superfamily", "family", "subfamily", "tribe", "subtribe", "genus", "subgenus", "section",
                 "subsection", "series", "species group", "species subgroup", "species", "subspecies", "varietas",
                 "forma", "strain"]

    def __str__(self):
        return "ID: %s Sci_Name: %s (%s)" % (self.tax_id, self.sci_name, self.rank)

    def get_lineage(self, taxon_dict):
        """
        Get all lineage information from root to given taxon
        :param taxon_dict: The output from func "build_taxon_database"
        :return: a lineage list from root to given taxon
        """
        lineage_list = [(self.tax_id, self.rank)]
        iter_taxon = self
        while not iter_taxon.tax_id == iter_taxon.parent_tax_id:
            iter_taxon = taxon_dict[iter_taxon.parent_tax_id]
            lineage_list.append((iter_taxon.tax_id, iter_taxon.rank))

        return lineage_list[::-1]

    def check_rank(self, tax_record_dict):
        from lib.xuyuxing.math.set_operating import uniqify
        from lib.xuyuxing.math.lcs import lcs
        for i in tax_record_dict:
            temp_list = tuple(
                uniqify([j[1] for j in tax_record_dict[i].get_lineage(tax_record_dict) if
                         j[1] != 'no rank' and j[1] != 'clade']))
            LCS_list = tuple(lcs(temp_list, Taxon.rank_list)[0])
            if not temp_list == LCS_list:
                print(temp_list)
                raise ValueError

    def lower_than(self, rank, tax_record_dict):
        temp_list = tuple([j[1] for j in self.get_lineage(tax_record_dict) if j[1] != 'no rank' and j[1] != 'clade'])

        if len(temp_list) == 0:  # Other sequence
            return False

        max_index = max(Taxon.rank_list.index(i) for i in temp_list)

        if max_index >= Taxon.rank_list.index(rank):
            return True
        else:
            return False

    def is_child_of(self, tax_id, tax_record_dict=None):
        if isinstance(self.get_lineage,list):
            parent_list = self.get_lineage
        else:
            parent_list = self.get_lineage(tax_record_dict)

        if tax_id in [i[0] for i in parent_list]:
            return True
        else:
            return False

    def is_ancestor_of(self, tax_id, tax_record_dict):
        taget_tax = tax_record_dict[tax_id]
        if isinstance(taget_tax.get_lineage,list):
            parent_list = tax_record_dict[tax_id].get_lineage
        else:
            parent_list = tax_record_dict[tax_id].get_lineage(tax_record_dict)
        if self.tax_id in [i[0] for i in parent_list]:
            return True
        else:
            return False

    def get_sons(self, tax_record_dict):
        sons_id_list = []
        for i in tax_record_dict:
            i = tax_record_dict[i]
            if self.tax_id == i.parent_tax_id:
                sons_id_list.append(i.tax_id)
        return sons_id_list


def build_taxon_database(taxonomy_dir, full_name_class=False):
    """
    NCBI taxonomy database can be download from: ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz
    In the dir a "names.dmp" and a "nodes.dmp" can be found, this func work for parse this two file to fit the class Ta\
    xon.
    :param taxonomy_dir: file path for uncompressed taxdump.tar.gz
    :return: a dict whose key is tax_id and value is a Taxon with attributes named "tax_id", "sci_name", "parent_tax_id\
    ", "rank" and "lineage" (Big memory)
    """
    names_dmp_file = taxonomy_dir + "/names.dmp"
    nodes_dmp_file = taxonomy_dir + "/nodes.dmp"

    tax_record = {}
    with open(names_dmp_file, 'r') as f:
        for each_line in f.readlines():
            each_line = each_line.strip()
            info = each_line.split("|")
            tax_id = re.sub('\t', '', info[0])
            name = re.sub('\t', '', info[1])
            # unique_name = re.sub('\t', '', info[2])
            name_class = re.sub('\t', '', info[3])

            if not tax_id in tax_record:
                tax_record[tax_id] = Taxon()
                tax_record[tax_id].tax_id = tax_id

            if name_class == "scientific name":
                tax_record[tax_id].sci_name = name
            elif full_name_class:
                if not hasattr(tax_record[tax_id], 'other_name'):
                    tax_record[tax_id].other_name = []
                tax_record[tax_id].other_name.append(name)

    with open(nodes_dmp_file, 'r') as f:
        for each_line in f.readlines():
            each_line = each_line.strip()
            info = each_line.split("|")
            tax_id = re.sub('\t', '', info[0])
            parent_tax_id = re.sub('\t', '', info[1])
            rank = re.sub('\t', '', info[2])
            tax_record[tax_id].parent_tax_id = parent_tax_id
            tax_record[tax_id].rank = rank

    for tax_id in tax_record:
        taxon = tax_record[tax_id]
        lineage_list = taxon.get_lineage(tax_record)
        # taxon.lineage = lineage_list
        for i in lineage_list:
            tax_id_tmp, rank_tmp = i
            if not (rank_tmp == 'no rank' or rank_tmp == 'clade'):
                setattr(taxon, rank_tmp, tax_id_tmp)

    merged_dmp = taxonomy_dir + "/merged.dmp"

    merged_record = {}
    with open(merged_dmp, 'r') as f:
        for each_line in f.readlines():
            each_line = each_line.strip()
            info = each_line.split("|")
            merged_id = re.sub('\t', '', info[0])
            new_id = re.sub('\t', '', info[1])
            merged_record[merged_id] = new_id

    for i in merged_record:
        tax_record[i] = copy.deepcopy(tax_record[merged_record[i]])
        tax_record[i].older = True

    return tax_record


def get_MRCA_taxon_id(taxon_id_list, tax_record_dict):
    if len(taxon_id_list) == 1:
        return taxon_id_list[0]

    lineage_sum_list = []
    for tax_id in taxon_id_list:
        tax_tmp = tax_record_dict[tax_id]
        lineage_list = tax_tmp.get_lineage(tax_record_dict)
        lineage_sum_list.append(lineage_list)

    range_num = min([len(i) for i in lineage_sum_list])

    MRCA = None
    for i in range(0, range_num):
        tmp_taxon_list = [j[i] for j in lineage_sum_list]
        if len(set(tmp_taxon_list)) == 1:
            MRCA = tmp_taxon_list[0]

    return MRCA[0]


def store_tax_record_into_sqlite(tax_record_dict, db_file):
    import time
    from lib.xuyuxing.base.common_command import log_print
    log_print("\t\tBegin: convert NCBI taxonomy database to sqlite3 file")
    start_time = time.time()
    column_name = ["tax_id", "sci_name", "parent_tax_id", "rank", "merged_new_id", "lineage"]
    sc.init_sql_db(db_file, "tax_record", column_name)
    num = 0
    record_tmp_dict = []
    for tax_id in tax_record_dict:
        record_tmp = []
        record_tmp.append(tax_id)
        taxon_tmp = tax_record_dict[tax_id]

        if hasattr(taxon_tmp, "sci_name"):
            record_tmp.append(getattr(taxon_tmp, "sci_name").replace("\"", ""))
        else:
            record_tmp.append("")

        if hasattr(taxon_tmp, "parent_tax_id"):
            record_tmp.append(getattr(taxon_tmp, "parent_tax_id"))
        else:
            record_tmp.append("")

        if hasattr(taxon_tmp, "rank"):
            record_tmp.append(getattr(taxon_tmp, "rank"))
        else:
            record_tmp.append("")

        if hasattr(taxon_tmp, "older"):
            record_tmp.append(getattr(taxon_tmp, "tax_id"))
        else:
            record_tmp.append("")

        lineage_list = taxon_tmp.get_lineage(tax_record_dict)
        lineage_string = ""
        for tax_id_rank, rank in lineage_list:
            lineage_string = lineage_string + "%s:%s;" % (tax_id_rank, rank)
        lineage_string = lineage_string.rstrip(";")
        record_tmp.append(lineage_string)

        num = num + 1
        record_tmp_dict.append(record_tmp)
        # sc.insert_one_record_to_sql_table(tuple(record_tmp), tuple(column_name), db_file, "tax_record")

        if num % 10000 == 0:
            sc.sqlite_write(record_tmp_dict, db_file, "tax_record", column_name)
            record_tmp_dict = []

        round_time = time.time()
        if round_time - start_time > 10:
            log_print("\t\t\t%d/%d" % (num, len(tax_record_dict)))
            start_time = round_time

    if len(record_tmp_dict) > 0:
        sc.sqlite_write(record_tmp_dict, db_file, "tax_record", column_name)
        record_tmp_dict = []
        log_print("\t\t\t%d/%d" % (num, len(tax_record_dict)))

    conn = sqlite3.connect(db_file)
    conn.execute("CREATE UNIQUE INDEX tax_id_index on tax_record (tax_id)")
    conn.close()

    log_print("\t\tEnd: convert NCBI taxonomy database to sqlite3 file")


class Taxon_db(Taxon):
    def lower_than(self, rank):
        temp_list = tuple([j[1] for j in self.get_lineage if j[1] != 'no rank' and j[1] != 'clade'])

        if len(temp_list) == 0:  # Other sequence
            return False

        max_index = max(Taxon.rank_list.index(i) for i in temp_list)

        if max_index >= Taxon.rank_list.index(rank):
            return True
        else:
            return False

    def is_a(self, taxon_id):
        if taxon_id in [i[0] for i in self.get_lineage]:
            return True
        else:
            return False


def read_tax_record_dict_db(db_file, tax_id_list=[]):
    column_name = ["tax_id", "sci_name", "parent_tax_id", "rank", "merged_new_id", "lineage"]

    record_tuple_list = sc.sqlite_select_by_a_key(db_file, "tax_record", "tax_id", tuple(tax_id_list))

    record_dict = {}
    for record_tuple in record_tuple_list:
        record = Taxon_db()
        record.tax_id = record_tuple[0]
        if not record_tuple[1] == "":
            record.sci_name = record_tuple[1]
        if not record_tuple[2] == "":
            record.parent_tax_id = record_tuple[2]
        if not record_tuple[3] == "":
            record.rank = record_tuple[3]
        if not record_tuple[4] == "":
            record.tax_id = record_tuple[4]
            record.older = True

        record.get_lineage = [tuple(i.split(":")) for i in record_tuple[5].split(";")]

        lineage_list = record.get_lineage
        # taxon.lineage = lineage_list
        for i in lineage_list:
            tax_id_tmp, rank_tmp = i
            if not (rank_tmp == 'no rank' or rank_tmp == 'clade'):
                setattr(record, rank_tmp, tax_id_tmp)
        record_dict[record_tuple[0]] = record

    return record_dict


def read_tax_name_record_dict_db(db_file, tax_name_list):
    column_name = ["tax_id", "sci_name", "parent_tax_id", "rank", "merged_new_id", "lineage"]

    record_tuple_list = sc.sqlite_select_by_a_key(db_file, "tax_record", "sci_name", tuple(tax_name_list))

    record_dict = {}
    for record_tuple in record_tuple_list:
        record = Taxon_db()
        record.tax_id = record_tuple[0]
        if not record_tuple[1] == "":
            record.sci_name = record_tuple[1]
        if not record_tuple[2] == "":
            record.parent_tax_id = record_tuple[2]
        if not record_tuple[3] == "":
            record.rank = record_tuple[3]
        if not record_tuple[4] == "":
            record.tax_id = record_tuple[4]
            record.older = True

        record.get_lineage = [tuple(i.split(":")) for i in record_tuple[5].split(";")]

        lineage_list = record.get_lineage
        # taxon.lineage = lineage_list
        for i in lineage_list:
            tax_id_tmp, rank_tmp = i
            if not (rank_tmp == 'no rank' or rank_tmp == 'clade'):
                setattr(record, rank_tmp, tax_id_tmp)
        record_dict[record_tuple[1]] = record

    return record_dict


def tax_id_list(db_file):
    conn = sqlite3.connect(db_file)
    record_tuple = [i[0] for i in conn.execute("select tax_id from tax_record").fetchall()]
    conn.close()
    return record_tuple


def get_common_tree(taxon_id_list, taxon_dict):
    MRCA = get_MRCA_taxon_id(taxon_id_list, taxon_dict)
    clade_MRCA = Clade(branch_length=None, name=MRCA, clades=None, confidence=None, color=None, width=None)
    clade_dict = {
        MRCA: clade_MRCA
    }

    for taxon_id in taxon_id_list:
        lineage = [i[0] for i in taxon_dict[taxon_id].get_lineage(taxon_dict)]

        for taxon_id_tmp in lineage:
            taxon_id_tmp_index = lineage.index(taxon_id_tmp)
            # old than MRCA
            if taxon_id_tmp_index < lineage.index(MRCA):
                continue
            # leaf taxon
            if taxon_id_tmp_index + 1 >= len(lineage):
                continue
            # good now add son to dict
            my_clade = clade_dict[taxon_id_tmp]
            son_taxon_id = lineage[taxon_id_tmp_index + 1]
            if my_clade.clades is None:
                my_clade.clades = []
            already_have_son = [i.name for i in my_clade.clades]
            if not son_taxon_id in already_have_son:
                son_clade = Clade(name=son_taxon_id)
                my_clade.clades.append(son_clade)
                clade_dict[son_taxon_id] = son_clade

    tree = Tree.from_clade(clade_MRCA)

    return tree


def blast_taxon_assign(taxon_score_list, common_tree, silence=True):
    """
    Give me blast score results of a seq, I will tell you the taxon of this sequence
    :param taxon_blast_score_dict: [("3702",100.00)]
    :param taxon_dict: build_taxon_database(taxon_dir)
    :return:
    """
    if not silence:
        print("# begin analysis:")

    # get the highest score for each taxon
    taxon_hash = {}
    for taxon_id, score in taxon_score_list:
        if not taxon_id in taxon_hash or taxon_hash[taxon_id] <= score:
            taxon_hash[taxon_id] = score

    if not silence:
        print("## get %d taxon\n" % len(taxon_hash))

    # build common_tree
    if not silence:
        print("# build common_tree:")

    taxon_id_list = list(taxon_hash.keys())
    # common_tree = get_common_tree(taxon_id_list, taxon_dict)
    common_tree = remove_pass_node(common_tree)
    common_tree = add_clade_name(common_tree)

    MRCA = common_tree.root.name
    if not silence:
        print("## get common_tree, MRCA: %s\n" % taxon_dict[MRCA].sci_name)

    leaf = common_tree.get_terminals()

    tree_dict = lookup_by_names(common_tree)

    # based on rank and fisher exact test
    if not silence:
        print("# based on rank and fisher exact test:")

    taxon_score_rank = sorted(taxon_hash, key=lambda x: taxon_hash[x], reverse=True)
    taxon_score_rank = [i for i in taxon_score_rank if i in tree_dict]

    fisher_p_dict = {}

    for clade in common_tree.find_clades(order='preorder'):
        all_sub_clade_id = [i.name for i in get_offspring(clade) + [clade]]
        all_sub_clade_have_hit = list(set(all_sub_clade_id) & set(taxon_id_list))

        top_hit_num = 0
        for i in all_sub_clade_have_hit:
            if taxon_score_rank.index(i) + 1 <= len(all_sub_clade_have_hit):
                top_hit_num += 1

        a = top_hit_num
        b = len(all_sub_clade_have_hit) - a
        c = b
        d = len(taxon_score_rank) - b

        oddsratio, pvalue = stats.fisher_exact([[a, b], [c, d]])
        clade.confidence = pvalue
        # print(pvalue)

        fisher_p_dict[clade.name] = pvalue

    fisher_qvalue_dict = get_qvalue(fisher_p_dict)

    good_clade_name = [i for i in sorted(fisher_qvalue_dict, key=lambda x: fisher_qvalue_dict[x]) if
                       fisher_qvalue_dict[i] <= 0.05]

    if not silence:
        print("# There are %d taxon have q_value <= 0.05" % len(good_clade_name))

    # based on score value to say how
    # compared between brother score
    if not silence:
        permutation_p_value = {}
        for clade_name in good_clade_name:
            clade_tmp = tree_dict[clade_name]
            parent_tmp = get_parent(clade_tmp, common_tree)

            if len(parent_tmp) != 0:
                clade_leaf = clade_tmp.get_terminals()
                brother_leaf = [i for i in parent_tmp.get_terminals() if i not in clade_leaf]

                clade_score = [taxon_hash[i.name] for i in clade_leaf if i.name in taxon_hash]
                brother_score = [taxon_hash[i.name] for i in brother_leaf if i.name in taxon_hash]

                p_value = permutation_test(clade_score, brother_score,
                                           method='approximate',
                                           func=lambda x, y: np.mean(x) - np.mean(y),
                                           seed=0)

                permutation_p_value[clade_name] = p_value

            else:
                permutation_p_value[clade_name] = 0

        print("Taxon_SciName\tTaxon_id\tFisher\tFDR\tPermutation")
        for i in good_clade_name:
            print("%s\t%s\t%.5e\t%.5e\t%.5f" % (
                taxon_dict[i].sci_name, i, fisher_p_dict[i], fisher_qvalue_dict[i], permutation_p_value[i]))
        # print(taxon_dict[i].sci_name, fisher_p_dict[i], fisher_qvalue_dict[i], permutation_p_value[i])

    # get full_supported_list
    if not silence:
        print("\n# based on node to find fully supported lineage:\n")

    full_supported_list = []
    for clade_name in good_clade_name:
        lineage_list = [i.name for i in get_ancestors(clade_name, common_tree)] + [clade_name]
        if set(lineage_list) & set(good_clade_name) == set(lineage_list):
            full_supported_list.append(clade_name)

    # full_supported_list = good_clade_name

    # get best lineage
    good_clade_lineage_list = []
    for clade_name in full_supported_list:
        lineage_list = [i.name for i in get_ancestors(clade_name, common_tree)] + [clade_name]
        good_clade_lineage_list.append(lineage_list)

    # if have two good lineage, use vote way
    merge_set_list = merge_subset(good_clade_lineage_list)

    vote_dict = {tuple(i): 0 for i in good_clade_lineage_list if set(i) in merge_set_list}

    for i in good_clade_lineage_list:
        for j in vote_dict:
            if set(i) & set(j) == set(i):
                vote_dict[j] += 1

    if not silence:
        num = 0
        for i in sorted(vote_dict, key=lambda x: vote_dict[x], reverse=True):
            print("#( %d ):" % num)
            lineage_txt = printer_list([taxon_dict[j].sci_name for j in i], head="\t", sep=";")
            print(lineage_txt)
            print("Vote num: %d\n" % (vote_dict[i]))
            num = num + 1

    good_lineage_list = list(vote_dict.keys())

    # use score compare
    # find conflict point
    # get full_supported_list
    if not silence:
        print("\n# get final results:")

    while len(good_lineage_list) > 1:
        for comp_pair in combinations(good_lineage_list, 2):
            if not silence:
                print("\tcompare with %s vs %s:" % (comp_pair[0][-1], comp_pair[1][-1]))

            A_clade = tree_dict[comp_pair[0][-1]]
            B_clade = tree_dict[comp_pair[1][-1]]
            mrca_clade = get_MRCA(A_clade, B_clade, common_tree)
            A_conflict_clade = [i for i in get_sons(mrca_clade) if i.name in comp_pair[0]][0]
            B_conflict_clade = [i for i in get_sons(mrca_clade) if i.name in comp_pair[1]][0]

            A_conflict_score = [taxon_hash[i.name] for i in A_conflict_clade.get_terminals() if
                                i.name in taxon_hash]
            B_conflict_score = [taxon_hash[i.name] for i in B_conflict_clade.get_terminals() if
                                i.name in taxon_hash]

            p_value = permutation_test(A_conflict_score, B_conflict_score,
                                       method='approximate',
                                       func=lambda x, y: np.mean(x) - np.mean(y),
                                       seed=0)
            if not silence:
                print("\tp_value: %f" % p_value)

            # A > B
            if p_value <= 0.05:
                good_lineage_list.remove(comp_pair[1])
                if not silence:
                    print("\tremove %s" % comp_pair[1][-1])
            # A < B
            elif p_value >= 0.95:
                good_lineage_list.remove(comp_pair[0])
                if not silence:
                    print("\tremove %s" % comp_pair[0][-1])
            else:
                good_lineage_list.remove(comp_pair[0])
                good_lineage_list.remove(comp_pair[1])

                mrca_lineage = tuple(
                    [i.name for i in get_ancestors(mrca_clade.name, common_tree)] + [mrca_clade.name])
                good_lineage_list.append(mrca_lineage)
                if not silence:
                    print("\tremove %s and %s, change to %s\n" % (comp_pair[0][-1], comp_pair[1][-1], mrca_clade.name))

            merge_set_list = merge_subset(good_lineage_list)

            good_lineage_list = [tuple(i) for i in good_clade_lineage_list if set(i) in merge_set_list]

            break

    if not silence:
        lineage_txt = printer_list([taxon_dict[j].sci_name for j in good_lineage_list[0]],
                                   head="Results:\t", sep="; ")

        print("\n" + lineage_txt + "\n")

        rename_dict = {}
        for i in tree_dict:
            sci_name = taxon_dict[i].sci_name
            if tree_dict[i].is_terminal() and i in taxon_hash:
                sci_name = sci_name + ": %.1f" % taxon_hash[i]

            rename_dict[i] = sci_name

        draw_ascii(common_tree, column_width=160, clade_name=True, rename_dict=rename_dict)

    if good_lineage_list == []:
        recommand_taxon = MRCA
        recommand_taxon_fisher = 1.0
    else:
        recommand_taxon = good_lineage_list[0][-1]
        recommand_taxon_fisher = fisher_qvalue_dict[good_lineage_list[0][-1]]

    return MRCA, recommand_taxon, recommand_taxon_fisher
